###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def users_ids(*, stringify_ids=_ELIDE, cursor=_ELIDE):
    """
    Returns an array of numeric user ids the authenticating user has muted.

    :param stringify_ids: Many programming environments will not consume our
        ids due to their size. Provide this option to have ids returned as
        strings instead. Read more about Twitter IDs.

    :param cursor: Causes the list of IDs to be broken into pages of no more
        than 5000 IDs at a time. The number of IDs returned is not guaranteed
        to be 5000 as suspended users are filtered out. If no cursor is
        provided, a value of -1 will be assumed, which is the first “page.” The
        response from the API will include a previous_cursor and next_cursor to
        allow paging back and forth. See Using cursors to navigate collections
        for more information.
    """
    binding = {'stringify_ids': stringify_ids, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/mutes/users/ids.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:mutes',
                           'get-mutes-users-ids',
                           binding)


def users_list(*, cursor=_ELIDE, include_entities=_ELIDE, skip_status=_ELIDE):
    """
    Returns an array of

    :param cursor: Causes the list of IDs to be broken into pages of no more
        than 5000 IDs at a time. The number of IDs returned is not guaranteed
        to be 5000 as suspended users are filtered out. If no cursor is
        provided, a value of -1 will be assumed, which is the first “page.” The
        response from the API will include a previous_cursor and next_cursor to
        allow paging back and forth. See Using cursors to navigate collections
        for more information.

    :param include_entities: The entities node will not be included when set to
        false.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'cursor': cursor, 'include_entities': include_entities,
               'skip_status': skip_status}
    url = 'https://api.twitter.com/1.1/mutes/users/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:mutes',
                           'get-mutes-users-list',
                           binding)


def users_create(*, screen_name=_ELIDE, user_id=_ELIDE):
    """
    Mutes the user specified in the ID parameter for the authenticating user.

    :param screen_name: The screen name of the potentially muted user. Helpful
        for disambiguating when a valid screen name is also a user ID.

    :param user_id: The ID of the potentially muted user. Helpful for
        disambiguating when a valid user ID is also a valid screen name.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id}
    url = 'https://api.twitter.com/1.1/mutes/users/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:mutes',
                           'post-mutes-users-create',
                           binding)


def users_destroy(*, screen_name=_ELIDE, user_id=_ELIDE):
    """
    Un-mutes the user specified in the ID parameter for the authenticating
    user.

    :param screen_name: The screen name of the potentially muted user. Helpful
        for disambiguating when a valid screen name is also a user ID.

    :param user_id: The ID of the potentially muted user. Helpful for
        disambiguating when a valid user ID is also a valid screen name.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id}
    url = 'https://api.twitter.com/1.1/mutes/users/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:mutes',
                           'post-mutes-users-destroy',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/mutes/users/ids.json'] = 'https://dev.twitter.com/rest/reference/get/mutes/users/ids'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/mutes/users/list.json'] = 'https://dev.twitter.com/rest/reference/get/mutes/users/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/mutes/users/create.json'] = 'https://dev.twitter.com/rest/reference/post/mutes/users/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/mutes/users/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/mutes/users/destroy'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def placements(*, product_type=_ELIDE):
    """
    Retrieve valid

    :param product_type: Scope the response to just the valid placements for
        the specified product type. Type: enum Possible values: LIVE_TV_EVENT,
        MEDIA, PROMOTED_ACCOUNT, PROMOTED_TWEETS
    """
    binding = {'product_type': product_type}
    url = 'https://ads-api.twitter.com/2/line_items/placements'
    return _TwitterRequest('GET',
                           url,
                           'ads:line_items',
                           'get-line-items-placements',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/line_items/placements'] = 'https://dev.twitter.com/ads/reference/get/line_items/placements'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def webhooks(url):
    """
    Registers a new webhook URL for the given application context. The URL will
    be validated via CRC request before saving. In case the validation fails,
    an error is returned. Only one webhook URL can be registered to an
    application.
    """
    binding = {'url': url}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks.json'
    return _TwitterRequest('POST',
                           url,
                           'webhooks:account_activity',
                           'post-account-activity-webhooks',
                           binding)


def webhooks():
    """
    Returns all URLs and their statuses for the given app. Currently, only one
    webhook URL can be registered to an application.
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks.json'
    return _TwitterRequest('GET',
                           url,
                           'webhooks:account_activity',
                           'get-account-activity-webhooks',
                           binding)


def webhooks_by_webhook_id(webhook_id):
    """
    Removes the webhook from the provided application’s configuration. The
    webhook ID can be accessed by making a call to GET
    /1.1/account_activity/webhooks.
    """
    binding = {'webhook_id': webhook_id}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}.json'
    url = url.format(**binding)
    return _TwitterRequest('DELETE',
                           url,
                           'webhooks:account_activity',
                           'delete-account-activity-webhooks-webhook-id',
                           binding)


def webhooks_by_webhook_id(webhook_id):
    """
    Triggers the challenge response check (CRC) for the given webhook’s URL. If
    the check is successful, returns 204 and reenables the webhook by setting
    its status to
    """
    binding = {'webhook_id': webhook_id}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}.json'
    url = url.format(**binding)
    return _TwitterRequest('PUT',
                           url,
                           'webhooks:account_activity',
                           'put-account-activity-webhooks-webhook-id',
                           binding)


def webhooks_subscriptions_by_webhook_id(webhook_id):
    """
    Subscribes the provided app to events for the provided user context. When
    subscribed, all DM events for the provided user will be sent to the app’s
    webhook via POST request.
    """
    binding = {'webhook_id': webhook_id}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}/subscriptions.json'
    url = url.format(**binding)
    return _TwitterRequest('POST',
                           url,
                           'webhooks:account_activity',
                           'post-account-activity-webhooks-webhook-id-subscriptions',
                           binding)


def webhooks_subscriptions_by_webhook_id(webhook_id):
    """
    Provides a way to determine if a webhook configuration is subscribed to the
    provided user’s Direct Messages. If the provided user context has an active
    subscription with the provided app, returns 204 OK. If the response code is
    not 204, then the user does not have an active subscription. See HTTP
    Response code and error messages below for details.
    """
    binding = {'webhook_id': webhook_id}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}/subscriptions.json'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'webhooks:account_activity',
                           'get-account-activity-webhooks-webhook-id-subscriptions',
                           binding)


def webhooks_subscriptions_by_webhook_id(webhook_id):
    """
    Deactivates subscription for the provided user context and app. After
    deactivation, all DM events for the requesting user will no longer be sent
    to the webhook URL.
    """
    binding = {'webhook_id': webhook_id}
    url = 'https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}/subscriptions.json'
    url = url.format(**binding)
    return _TwitterRequest('DELETE',
                           url,
                           'webhooks:account_activity',
                           'delete-account-activity-webhooks-webhook-id-subscriptions',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks.json'] = 'https://dev.twitter.com/webhooks/reference/post/account_activity/webhooks'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks.json'] = 'https://dev.twitter.com/webhooks/reference/get/account_activity/webhooks'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}.json'] = 'https://dev.twitter.com/webhooks/reference/del/account_activity/webhooks'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}.json'] = 'https://dev.twitter.com/webhooks/reference/put/account_activity/webhooks'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}/subscriptions.json'] = 'https://dev.twitter.com/webhooks/reference/post/account_activity/webhooks/subscriptions'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}/subscriptions.json'] = 'https://dev.twitter.com/webhooks/reference/get/account_activity/webhooks/subscriptions'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/account_activity/webhooks/{webhook_id}/subscriptions.json'] = 'https://dev.twitter.com/webhooks/reference/del/account_activity/webhooks/subscriptions'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def id_by_place_id(place_id):
    """
    Returns all the information about a known

    :param place_id: A place in the world. These IDs can be retrieved from
        geo/reverse_geocode.
    """
    binding = {'place_id': place_id}
    url = 'https://api.twitter.com/1.1/geo/id/{place_id}.json'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'rest:geo',
                           'get-geo-id-place-id',
                           binding)


def reverse_code(lat, long, *, accuracy=_ELIDE, granularity=_ELIDE,
                 max_results=_ELIDE, callback=_ELIDE):
    """
    Given a latitude and a longitude, searches for up to 20 places that can be
    used as a

    :param lat: The latitude to search around. This parameter will be ignored
        unless it is inside the range -90.0 to +90.0 (North is positive)
        inclusive. It will also be ignored if there isn’t a corresponding long
        parameter.

    :param long: The longitude to search around. The valid ranges for longitude
        is -180.0 to +180.0 (East is positive) inclusive. This parameter will
        be ignored if outside that range, if it is not a number, if geo_enabled
        is disabled, or if there not a corresponding lat parameter.

    :param accuracy: A hint on the “region” in which to search. If a number,
        then this is a radius in meters, but it can also take a string that is
        suffixed with ft to specify feet. If this is not passed in, then it is
        assumed to be 0m. If coming from a device, in practice, this value is
        whatever accuracy the device has measuring its location (whether it be
        coming from a GPS, WiFi triangulation, etc.).

    :param granularity: This is the minimal granularity of place types to
        return and must be one of: poi, neighborhood, city, admin or country.
        If no granularity is provided for the request neighborhood is assumed.
        Setting this to city, for example, will find places which have a type
        of city, admin or country.

    :param max_results: A hint as to the number of results to return. This does
        not guarantee that the number of results returned will equal
        max_results, but instead informs how many “nearby” results to return.
        Ideally, only pass in the number of places you intend to display to the
        user here.

    :param callback: If supplied, the response will use the JSONP format with a
        callback of the given name.
    """
    binding = {'lat': lat, 'long': long, 'accuracy': accuracy, 'granularity':
               granularity, 'max_results': max_results, 'callback': callback}
    url = 'https://api.twitter.com/1.1/geo/reverse_geocode.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:geo',
                           'get-geo-reverse-geocode',
                           binding)


def search(*, lat=_ELIDE, long=_ELIDE, query=_ELIDE, ip=_ELIDE,
           granularity=_ELIDE, accuracy=_ELIDE, max_results=_ELIDE,
           contained_within=_ELIDE, attribute_street_address=_ELIDE,
           callback=_ELIDE):
    """
    Search for places that can be attached to a statuses/update. Given a
    latitude and a longitude pair, an IP address, or a name, this request will
    return a list of all the valid places that can be used as the

    :param lat: The latitude to search around. This parameter will be ignored
        unless it is inside the range -90.0 to +90.0 (North is positive)
        inclusive. It will also be ignored if there isn’t a corresponding long
        parameter.

    :param long: The longitude to search around. The valid ranges for longitude
        is -180.0 to +180.0 (East is positive) inclusive. This parameter will
        be ignored if outside that range, if it is not a number, if geo_enabled
        is disabled, or if there not a corresponding lat parameter.

    :param query: Free-form text to match against while executing a geo-based
        query, best suited for finding nearby locations by name. Remember to
        URL encode the query.

    :param ip: An IP address. Used when attempting to fix geolocation based off
        of the user’s IP address.

    :param granularity: This is the minimal granularity of place types to
        return and must be one of: poi, neighborhood, city, admin or country.
        If no granularity is provided for the request neighborhood is assumed.
        Setting this to city, for example, will find places which have a type
        of city, admin or country.

    :param accuracy: A hint on the “region” in which to search. If a number,
        then this is a radius in meters, but it can also take a string that is
        suffixed with ft to specify feet. If this is not passed in, then it is
        assumed to be 0m. If coming from a device, in practice, this value is
        whatever accuracy the device has measuring its location (whether it be
        coming from a GPS, WiFi triangulation, etc.).

    :param max_results: A hint as to the number of results to return. This does
        not guarantee that the number of results returned will equal
        max_results, but instead informs how many “nearby” results to return.
        Ideally, only pass in the number of places you intend to display to the
        user here.

    :param contained_within: This is the place_id which you would like to
        restrict the search results to. Setting this value means only places
        within the given place_id will be found. Specify a place_id. For
        example, to scope all results to places within “San Francisco, CA USA”,
        you would specify a place_id of “5a110d312052166f”

    :param attribute_street_address: This parameter searches for places which
        have this given street address. There are other well-known, and
        application specific attributes available. Custom attributes are also
        permitted. Learn more about [node:208, title=”Place Attributes”].

    :param callback: If supplied, the response will use the JSONP format with a
        callback of the given name.
    """
    binding = {'lat': lat, 'long': long, 'query': query, 'ip': ip,
               'granularity': granularity, 'accuracy': accuracy,
               'max_results': max_results, 'contained_within':
               contained_within, 'attribute:street_address':
               attribute_street_address, 'callback': callback}
    url = 'https://api.twitter.com/1.1/geo/search.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:geo',
                           'get-geo-search',
                           binding)


def place(name, contained_within, token, lat, long, *,
          attribute_street_address=_ELIDE, callback=_ELIDE):
    """
    As of December 2nd, 2013, this endpoint

    :param name: The name a place is known as.

    :param contained_within: The place_id within which the new place can be
        found. Try and be as close as possible with the containing place. For
        example, for a room in a building, set the contained_within as the
        building place_id.

    :param token: The token found in the response from geo/similar_places.

    :param lat: The latitude the place is located at. This parameter will be
        ignored unless it is inside the range -90.0 to +90.0 (North is
        positive) inclusive. It will also be ignored if there isn’t a
        corresponding long parameter.

    :param long: The longitude the place is located at. The valid ranges for
        longitude is -180.0 to +180.0 (East is positive) inclusive. This
        parameter will be ignored if outside that range, if it is not a number,
        if geo_enabled is disabled, or if there not a corresponding lat
        parameter.

    :param attribute_street_address: This parameter searches for places which
        have this given street address. There are other well-known, and
        application specific attributes available. Custom attributes are also
        permitted. Learn more here.

    :param callback: If supplied, the response will use the JSONP format with a
        callback of the given name.
    """
    binding = {'name': name, 'contained_within': contained_within, 'token':
               token, 'lat': lat, 'long': long, 'attribute:street_address':
               attribute_street_address, 'callback': callback}
    url = 'https://api.twitter.com/1.1/geo/place.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:geo',
                           'post-geo-place',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/geo/id/{place_id}.json'] = 'https://dev.twitter.com/rest/reference/get/geo/id/place_id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/geo/reverse_geocode.json'] = 'https://dev.twitter.com/rest/reference/get/geo/reverse_geocode'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/geo/search.json'] = 'https://dev.twitter.com/rest/reference/get/geo/search'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/geo/place.json'] = 'https://dev.twitter.com/rest/reference/post/geo/place'



import os
import pytest
import brittle_wit as bw
import brittle_wit.rest_api as rest_api
from tests.helpers import load_fixture_txt, AsyncSafeTestCase
from brittle_wit_core import TwitterError
from brittle_wit.helpers import (parse_datetime, grouping, running_task_names,
                                 linear_backoff, expon_backoff, retrying,
                                 _twitter_request_html_repr, CircularQueue)


def failure_script(ret_val, *failure_schedule):
    failure_schedule = list(failure_schedule)

    async def _f():
        if not failure_schedule:
            return ret_val
        else:
            raise failure_schedule.pop(0)
    return _f


def make_twitter_error(status_code, mocker):
    resp = mocker.MagicMock()
    resp.status = status_code
    return TwitterError(None, None, resp, "none")


def test_parse_datetime():
    s = "tue mar 29 15:40:03 +0000 2016"
    assert parse_datetime(s).timestamp() == 1459266003.0


def test_grouping():
    items = list(range(7))
    expected = [[0, 1, 2], [3, 4, 5], [6]]
    assert list(grouping(items, 3)) == expected


def test_linear_backoff():
    assert linear_backoff(3, 60) == 180


def test_expon_backoff():
    assert expon_backoff(3, 60) == 240


@pytest.mark.asyncio
async def test_retrying_on_no_failures():
    coro = failure_script(True)
    assert await retrying(coro, 1)


@pytest.mark.asyncio
async def test_retrying_non_retryable(mocker):
    err = make_twitter_error(1, mocker)
    coro = failure_script(True, err)
    with pytest.raises(TwitterError):
        assert await retrying(coro, 2)


@pytest.mark.asyncio
async def test_retrying_after_one_failure_tries_left(mocker):
    err = make_twitter_error(500, mocker)
    coro = failure_script(True, err)
    assert await retrying(coro, 2, sleep_time=0)


@pytest.mark.asyncio
async def test_retrying_after_one_failure_after_all_tries(mocker):
    err = make_twitter_error(500, mocker)
    coro = failure_script(True, err)
    with pytest.raises(TwitterError):
        assert await retrying(coro, 1)


@pytest.mark.asyncio
async def test_retrying_given_known_retryable_codes(mocker):
    # Be overly cautious with this test. It's hard to test services!
    for code in [500, 502, 503, 504]:
        err = make_twitter_error(code, mocker)
        coro = failure_script(True, err)
        with pytest.raises(TwitterError):
            assert await retrying(coro, 1)


@pytest.mark.asyncio
async def test_running_task_names():
    assert ['test_running_task_names'] == running_task_names()



import pytest
from asyncio import TimeoutError
from brittle_wit_core import (AppCredentials,
                              ClientCredentials, TwitterRequest,
                              TwitterError)
from brittle_wit.streaming.twitter_stream import TwitterStream
from tests.helpers import mock_http_req, mock_content_reader


pytestmark = pytest.mark.asyncio
app_cred = AppCredentials("app", "secret")
client_cred = ClientCredentials(1, "client", "secret")


@pytest.fixture
def stream(mocker):
    return TwitterStream(mocker.MagicMock(),
                         app_cred,
                         client_cred,
                         mocker.MagicMock())


async def test_is_not_open_after_instantiation(stream, mocker):
    assert not stream.is_open()


async def test_no_double_connecting(stream, mocker):
    mocker.patch.object(stream, 'is_open', return_value=True)
    with pytest.raises(RuntimeError):
        stream._connect()


async def test_disconnect_closes_underlying_http_req(stream, mocker):
    http_req = mocker.MagicMock()
    stream._http_req = http_req
    stream._resp = 42

    stream.disconnect()

    assert http_req.close.called
    assert stream._http_req is None
    assert stream._resp is None


async def test_reconnect_purges_buffer_and_mailbox(stream, mocker):
    connect = mocker.patch.object(stream, '_connect')
    stream._entry_processor = mocker.MagicMock()

    stream.reconnect(clear_prior_messages=True)

    assert stream._entry_processor.purge_mailbox.called
    assert stream._entry_processor.purge_buffer.called


async def test_reconnect_forces_closure(stream, mocker):
    connect = mocker.patch.object(stream, '_connect')

    http_req = mocker.MagicMock()
    stream._http_req = http_req
    stream._resp = mocker.MagicMock()
    stream._resp.connection.closed = False

    stream.reconnect()

    assert http_req.close.called
    assert stream._http_req is None
    assert stream._resp is None

    connect.assert_called_once_with()


async def test_reconnect_errs_if_open(stream, mocker):
    connect = mocker.patch.object(stream, 'is_open', return_value=True)

    with pytest.raises(RuntimeError) as e:
        stream.reconnect(force_close_if_open=False)

    assert 'Already connected' in str(e)


async def test__aiter__connects_if_closed(stream, mocker):
    stream._resp = mocker.MagicMock()
    stream._resp.connection.closed = True
    stream._connect = mocker.MagicMock()
    stream._http_req = mock_http_req(200, False)

    await stream.__aiter__()
    stream._connect.assert_called_once_with()


async def test__aiter__gets_resp_if_none(stream, mocker):
    stream._resp = mocker.MagicMock()
    stream._resp.connection.closed = True
    stream._connect = mocker.MagicMock()
    stream._http_req = mock_http_req(200, False)

    await stream.__aiter__()
    stream._connect.assert_called_once_with()

    stream._resp = None
    assert await stream.__aiter__() is stream
    assert stream._resp is not None


async def test__aiter__300_status(stream, mocker):
    pred = mocker.patch.object(stream, 'is_open', return_value=True)

    stream._http_req = mock_http_req(500, True, 'mock text')

    with pytest.raises(TwitterError):
        await stream.__aiter__()

    pred.assert_called_once_with()


async def test__anext__(stream, mocker):
    messages = [b"this is a message\r\nthis is another\r\n"]

    stream._resp = mock_content_reader(messages)
    msg = await stream.__anext__()
    assert msg == b"this is a message"

    msg = await stream.__anext__()
    assert msg == b"this is another"

    with pytest.raises(StopAsyncIteration):
        msg = await stream.__anext__()


async def test_read_hang_timeout_for__anext__(stream):
    stream._read_hang_timeout = 0.001
    messages = [b"this is a message\r\nthis is another\r\n"]

    stream._resp = mock_content_reader(messages, delay=0.05)
    with pytest.raises(TimeoutError):
        msg = await stream.__anext__()



import pytest
import aiohttp
import asyncio
from unittest.mock import patch, MagicMock
from brittle_wit import AppCredentials, ClientCredentials
from brittle_wit.app import App, ClientContext, RUNNING, PRESTINE, FINISHED
from brittle_wit.executors import ManagedClientRequestProcessors


pytestmark = pytest.mark.asyncio


# XXX: This simplifies things a lot.
def async_return(result):
    f = asyncio.Future()
    f.set_result(result)
    return f


async def test_client_context(app, client_cred, mocker):
    ctx = ClientContext(app, client_cred)

    execute_path = 'brittle_wit.executors.ClientRequestProcessor.execute'
    with mocker.patch(execute_path, return_value=async_return(42)) as execute:
        assert await ctx.execute('my_request') == 42


async def test_properties(app, app_cred, client_cred):
    # =========================================================================
    # These are silly tests meant for boosting coverage. They protect very
    # little. But, people care about coverage badges. At least at first,
    # possibly fragile tests that don't indicate true failures is worth the
    # possibility of greater adoption due to high coverage.
    # =========================================================================
    assert app.app_cred == app_cred
    assert isinstance(app.session, aiohttp.ClientSession)
    assert isinstance(app.manager, ManagedClientRequestProcessors)
    assert app.client_context(client_cred)._client_cred == client_cred


async def the_answer_coro():
    return 42


async def test_schedule_when_prestine(app):
    assert len(app._initial_coros) == 0
    c = the_answer_coro()
    assert app.schedule(c) is None  # Not running so no returned task.
    assert len(app._initial_coros) == 1
    await c  # To clear warnings about failured to run coro.


async def test_schedule_when_finished(app):
    app._state = FINISHED  # I.e. not in {RUNNING, PRESTINE}
    c = the_answer_coro()
    with pytest.raises(RuntimeError) as e:
        app.schedule(c)
    assert "state finished" in str(e)
    await c  # To clear warnings about failured to run coro.


async def test_schedule_when_running(app):
    app._state = RUNNING
    task = app.schedule(the_answer_coro())
    assert isinstance(task, asyncio.Task)
    assert await task == 42


async def test_schedule_maintenance_when_prestine(app):
    app._state = PRESTINE
    c = the_answer_coro()
    app.schedule_maintenance(c, True)
    assert c in app._maintenance_coros
    assert app._maintenance_coros[c]  # i.e. requires cancelation
    await c  # To clear warnings about failured to run coro.


async def test_schedule_maintenance_when_running(app):
    app._state = RUNNING
    c = the_answer_coro()
    task = app.schedule_maintenance(c, True)

    assert isinstance(task, asyncio.Task)
    assert c in app._maintenance_coros
    assert app._maintenance_coros[c]  # i.e. requires cancelation
    assert await task == 42


async def test_schedule_maintenance_when_finished(app):
    app._state = FINISHED
    c = the_answer_coro()
    with pytest.raises(RuntimeError) as e:
        app.schedule_maintenance(c, True)
    assert "state finished" in str(e)
    await c  # To clear warnings about failured to run coro.


# XXX: Add tests for no send cancel.


async def test_run_async_after_finished(app):
    app._state = FINISHED
    with pytest.raises(RuntimeError) as e:
        await app.run_async()
    assert 'already finished' in str(e)


async def test_run_async_when_running(app):
    app._state = 'running'
    with pytest.raises(RuntimeError) as e:
        await app.run_async()
    assert 'already running' in str(e)



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def iab_categories(*, count=_ELIDE, cursor=_ELIDE, with_total_count=_ELIDE):
    """
    Request the valid app

    :param count: Specifies the number of records to try and retrieve per
        distinct request. Type: int Default: 200 Min, Max: 1, 1000

    :param cursor: Specifies a cursor to get the next page of categories. See
        Pagination for more information. Type: string Example: gc-ddf4a

    :param with_total_count: Include the total_count response attribute. Note:
        total_count will only be returned when with_total_count is set to true
        and next_cursor is null. Note: Requests which include total_count will
        have lower rate limits, currently set at 200 per 15 minutes. Type:
        boolean Default: false Possible values: true, false
    """
    binding = {'count': count, 'cursor': cursor, 'with_total_count':
               with_total_count}
    url = 'https://ads-api.twitter.com/2/iab_categories'
    return _TwitterRequest('GET',
                           url,
                           'ads:iab_categories',
                           'get-iab-categories',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/iab_categories'] = 'https://dev.twitter.com/ads/reference/get/iab_categories'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def tweets(q, *, geocode=_ELIDE, lang=_ELIDE, locale=_ELIDE,
           result_type=_ELIDE, count=_ELIDE, until=_ELIDE, since_id=_ELIDE,
           max_id=_ELIDE, include_entities=_ELIDE):
    """
    Returns a collection of relevant

    :param q: A UTF-8, URL-encoded search query of 500 characters maximum,
        including operators. Queries may additionally be limited by complexity.

    :param geocode: Returns tweets by users located within a given radius of
        the given latitude/longitude. The location is preferentially taking
        from the Geotagging API, but will fall back to their Twitter profile.
        The parameter value is specified by ” latitude,longitude,radius ”,
        where radius units must be specified as either ” mi ” (miles) or ” km ”
        (kilometers). Note that you cannot use the near operator via the API to
        geocode arbitrary locations; however you can use this geocode parameter
        to search near geocodes directly. A maximum of 1,000 distinct “sub-
        regions” will be considered when using the radius modifier.

    :param lang: Restricts tweets to the given language, given by an ISO 639-1
        code. Language detection is best-effort.

    :param locale: Specify the language of the query you are sending (only ja
        is currently effective). This is intended for language-specific
        consumers and the default should work in the majority of cases.

    :param result_type: Optional. Specifies what type of search results you
        would prefer to receive. The current default is “mixed.” Valid values
        include: * mixed : Include both popular and real time results in the
        response. * recent : return only the most recent results in the
        response * popular : return only the most popular results in the
        response.

    :param count: The number of tweets to return per page, up to a maximum of
        100. Defaults to 15. This was formerly the “rpp” parameter in the old
        Search API.

    :param until: Returns tweets created before the given date. Date should be
        formatted as YYYY-MM-DD. Keep in mind that the search index has a 7-day
        limit. In other words, no tweets will be found for a date older than
        one week.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param include_entities: The entities node will not be included when set to
        false.
    """
    binding = {'q': q, 'geocode': geocode, 'lang': lang, 'locale': locale,
               'result_type': result_type, 'count': count, 'until': until,
               'since_id': since_id, 'max_id': max_id, 'include_entities':
               include_entities}
    url = 'https://api.twitter.com/1.1/search/tweets.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:search',
                           'get-search-tweets',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/search/tweets.json'] = 'https://dev.twitter.com/rest/reference/get/search/tweets'



import asyncio
import logging
from collections import OrderedDict, deque
from datetime import datetime
from brittle_wit_core import TwitterRequest, WrappedException


LOGGER = logging.getLogger('brittle_wit')


def parse_datetime(s):
    """
    Parse a date string sent in a tweet.
    :param s: a Twitter-returned date-time string
    :return: a datetime object
    """
    return datetime.strptime(s, "%a %b %d %H:%M:%S %z %Y")


def linear_backoff(failure_num, sleep_time):
    return failure_num * sleep_time


def expon_backoff(failure_num, sleep_time):
    return 2**(failure_num - 1) * sleep_time


async def retrying(make_coro, n_tries, strategy=linear_backoff, sleep_time=60):
    for attempt in range(1, n_tries + 1):
        try:
            return await make_coro()
        except Exception as e:
            LOGGER.error("Error %s", e)
            wrapped = WrappedException.wrap_if_nessessary(e)
            if not wrapped.is_retryable or attempt == n_tries:
                raise wrapped
            await asyncio.sleep(strategy(attempt, sleep_time))


def grouping(items, n):
    """
    Group items into sequential chunks.

    Each chunk has a length of n. However, the last chunk is somewhere
    between 1 and n.

    :param items: the iterable of items
    :param n: the chunk size
    """
    group = []
    for item in items:
        if len(group) == n:
            yield group
            group = []
        group.append(item)

    if group:
        yield group


def _make_table(d, header_labels=None, sort=False):
    snippets = ["<table>"]

    if header_labels:
        assert len(header_labels) == 2
        snippets.append("<tr><th>{}</th><th>{}</th>".format(*header_labels))

    row_str = "<tr><td>{}</td><td>{}</td>"

    ks = sorted(d) if sort else d
    for k in ks:
        snippets.append(row_str.format(k, d[k]))

    snippets.append("</table>")

    return "\n".join(snippets)


def _twitter_request_html_repr(self):
    d = OrderedDict()

    for k in ['_method', '_url', '_family', '_service', '_parse_as']:
        d[k[1:]] = getattr(self, k)

    if self._params:
        d['Params'] = _make_table(self._params, ('Key', 'Value'), True)

    api_url = TwitterRequest.DOC_URLS.get(self._url, None)
    if api_url:
        api_link = "<a href='{}' target='_new'>{}</a>".format(api_url, api_url)
        d["docs"] = api_link.replace(".json", "")

    return _make_table(d, ("Name", "Value"))


def monkey_patch_ipython_disp():
    TwitterRequest._repr_html_ = _twitter_request_html_repr


def running_task_names():
    """
    :return: list of all ``qual_name``s for running tasks.
    """
    tasks = asyncio.Task.all_tasks()
    return [t._coro.__qualname__ for t in tasks if not t.done()]


class CircularQueue:
    """
    A Circular Queue with Fast Membership Testing.
    """

    def __init__(self):
        self._queue, self._members = deque(), set()

    def __contains__(self, item):
        return item in self._members

    def __bool__(self):
        return bool(self._queue)

    def __delitem__(self, item):
        self._members.remove(item)
        self._queue.remove(item)

    def add(self, item):
        """
        Adds an item to the FIRST slot in the queue.

        :raises: a ValueError if the item is already enqueued.
        :param item: the item to add.
        """
        if item in self._members:
            raise ValueError("Item {} already in queue".format(item))

        self._queue.appendleft(item)
        self._members.add(item)

    def pop_and_rotate(self):
        """
        :return: the next item on the queue while maintaining
            the circular buffer.
        """
        if not self._queue:
            return None

        item = self._queue.popleft()
        self._queue.append(item)
        return item



import os
import re
import textwrap
from collections import defaultdict


def pep8_join(tokens, sep, init_indent='', next_indent='    ', tail_k=0):
    """
    Join a sequence of tokens for PEP8-compliant code generation.

    :param tokens: an iterable of tokens
    :param sep: a string separator to delimit the tokens
    :param init_indent: a string for the initial line's indentation
    :param next_indent: a string for all subsequent lines' indentation
    :param tail_k: subtracted from 79 to get the maximum line length
    """
    return textwrap.fill(sep.join(str(tok) for tok in tokens),
                         width=79 - tail_k,
                         initial_indent=init_indent,
                         break_long_words=False,
                         subsequent_indent=next_indent,
                         drop_whitespace=True)


RESERVED_WORDS = {'False', 'None', 'True', 'and', 'as', 'assert', 'break',
                  'class', 'continue', 'def', 'del', 'elif', 'else', 'except',
                  'finally', 'for', 'from', 'global', 'if', 'import', 'in',
                  'is', 'lambda', 'nonlocal', 'not', 'or', 'pass', 'raise',
                  'return', 'try', 'while', 'with', 'yield'}


def _param_name_translations(param_names):
    """
    :return: a dict from the original name to the source-safe
        name.
    """
    translations = {}
    for k in param_names:
        name = k

        name = name.replace(':', '_').replace('[]', '')

        if name in RESERVED_WORDS:
            name += '_'

        assert name.isidentifier()

        translations[k] = name

    return translations


def _generate_def_line(api_def, method_name_in_fname=False, **renamings):
    name = _generate_func_name(api_def, method_name_in_fname)
    declaration = "def {}(".format(name)

    tokens = [renamings.get(k, k) + ("={}".format(default) if default else "")
              for k, default in _generate_param_tokens(api_def)]
    if not tokens:
        return declaration + "):"

    return pep8_join(tokens,
                     ", ",
                     init_indent=declaration,
                     next_indent=len(declaration) * " ",
                     tail_k=2) + "):"


def _generate_param_doc(param_def, name):
    """
    Generate the parameter part of a doc string

    The name is taken as an argument because some names like
    ``with`` are reserved words.

    :param param_def: the parameter dict from the api json.
    :param name: the parameter's name
    """
    desc, kind = param_def.get('description'), param_def.get('type')

    # There is no additional documentation for this argument, so
    # don't document it more than the implicit function defition.
    if not desc and not kind:
        return None

    desc = (desc or "") + (" ({})".format(kind) if kind else "")

    doc_str = ":param {}: {}".format(name, desc)

    return textwrap.fill(doc_str, width=79,
                         initial_indent="    ",
                         subsequent_indent="        ",
                         replace_whitespace=True,
                         break_long_words=False)


def _generate_doc_str(api_def, **renamings):
    desc_line = textwrap.fill(api_def['desc'] or "",
                              width=79,
                              initial_indent='    ',
                              subsequent_indent='    ',
                              replace_whitespace=True,
                              break_long_words=False)

    param_docs = []
    for param_def in api_def['params']:
        k = param_def['name']
        s = _generate_param_doc(param_def, renamings.get(k, k))
        if s:
            param_docs.append(s)
    param_doc = "\n\n" + "\n\n".join(param_docs) if param_docs else ''

    doc = "\n".join([desc_line]) + param_doc

    return '    """\n' + doc + '\n    """'


def _generate_binding(api_def, **renamings):
    s = "    binding = {"
    indent = len(s) * " "

    bindings = []
    for param in api_def['params']:
        name = param['name']
        binding = "'{}': {}".format(name, renamings.get(name, name))
        bindings.append(binding)
    if not bindings:
        return s + "}"

    bindings = pep8_join(bindings,
                         sep=", ",
                         init_indent=s,
                         next_indent=indent,
                         tail_k=1)

    return bindings + "}"


def _generate_url_block(api_def):
    lines = ["    url = '{}'".format(api_def['url'])]

    if api_def['slugs']:
        lines.append("    url = url.format(**binding)")

    return "\n".join(lines)


REQ_TEMPLATE = "    " + """
    return _TwitterRequest('{method}',
                           url,
                           '{family}',
                           '{service}',
                           binding)
""".strip()


def _generate_req_str(api_def):
    return REQ_TEMPLATE.format(**api_def)


def generate_source(api_def, use_http_method=False):
    param_names = [d['name'] for d in api_def['params']]
    renamings = _param_name_translations(param_names)

    def_line = _generate_def_line(api_def,
                                  method_name_in_fname=use_http_method,
                                  **renamings)
    doc_str = _generate_doc_str(api_def, **renamings)
    binding_str = _generate_binding(api_def, **renamings)
    url_str = _generate_url_block(api_def)
    ret_str = _generate_req_str(api_def)

    return "\n".join([def_line, doc_str, binding_str, url_str, ret_str])


AUTOGEN_HEADER = '''
###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################
'''.strip()


IMPORT_HEADER = """
from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE
""".strip()

def generate_modules(base_dir, definitions, func_gen=generate_source):
    """
    Generate the module file skeleton given the definitions.

    - For each group, generate ``{group}_api`` module
    - For each family, generate ``{group}_api.{family}`` module

    :param base_dir: the file directory in which to create the modules
    :param definitions: the extracted api definitions
    :param func_gen: a function that returns a string of generated code
        when given an api_def
    """
    paths = defaultdict(set)
    code_blocks = defaultdict(list)
    doc_refs = defaultdict(list)

    # Gather the paths from the definitions.
    for group, api_defs in definitions.items():
        for api_def in api_defs:
            family = api_def['family'].split(":")[1]
            paths[group].add(family)
            code_blocks[(group, family)].append(func_gen(api_def))
            if 'url' in api_def and 'reference_url' in api_def:
                doc_pair = (api_def['url'], api_def['reference_url'])
                doc_refs[(group, family)].append(doc_pair)

    # Create the skeleton.
    for group, families in paths.items():
        # Setup parent module.
        parent_module = group + "_api"
        parent_dir = os.path.join(base_dir, parent_module)
        os.makedirs(parent_dir, exist_ok=True)

        # Write all the definitions.
        for family in families:
            sub_module = os.path.join(parent_dir, family + ".py")
            with open(sub_module, "w") as fp:
                fp.write(AUTOGEN_HEADER + "\n\n\n")
                fp.write(IMPORT_HEADER + "\n\n\n")

                for code_block in code_blocks[group, family]:
                    fp.write(code_block + "\n\n\n")

                for url, ref_url in doc_refs[group, family]:
                    doc_line = "_TwitterRequest.DOC_URLS['{}'] = '{}'\n"
                    fp.write(doc_line.format(url, ref_url))

        import_fmt = "import brittle_wit.{}.{}\n"
        with open(os.path.join(parent_dir, '__init__.py'), "w") as fp:
            fp.write(AUTOGEN_HEADER + "\n\n")

            for family in families:
                fp.write(import_fmt.format(parent_module, family))


def _generate_param_tokens(api_def):
    """
    Generate the function's parameter tokens.

    :param api_def: an api definition.
    """
    params = api_def.get('params')

    if not params:
        return []

    # Slugs always come first; then required; then optional.
    slugged, required, optional = [], [], []
    for param in params:
        if param['required']:
            if param['name'] in api_def['slugs']:
                slugged.append(param)
            else:
                required.append(param)
        else:
            optional.append(param)

    # Required params come first.
    tokens = [(param['name'], None) for param in list(reversed(slugged)) + required]

    if optional:
        tokens.append(('*', None))

    # Optionals are keyword-only to protect fail loudly.
    for param in optional:
        tokens.append((param['name'], '_ELIDE'))

    return tokens


SLUGS_RE = re.compile(r":[A-Za-z_0-9]+")
SLUG_POSTFIX_RE = re.compile(r"^.*?(:[A-Za-z_0-9]+)$")
SLUG_INFIX_RE = re.compile(r"""^(?:[A-Za-z]*/)?
                               ([A-Za-z0-9_]+)/
                               (\:[A-Za-z_0-9]+)/
                               (.*)$""",
                           re.X)


def _simplify_path(path, family):
    # Special case for some advertising api calls.
    if path.startswith("1/"):
        path = path[2:]
    return path.replace(family, "")


def _partition_path(path, known_params, slugs):
    namespaces, placeholders = [], []

    for part in path.split("/"):
        if part.startswith(':'):
            k = part[1:]
            assert k in known_params, k
            placeholders.append(k)
            assert k in slugs
        else:
            namespaces.append(part)

    return namespaces, placeholders


def _generate_func_name(api_def, append_http_method=False):
    """
    Derive a function name from an api_def

    :param append_http_method: if True, then the method name will end
        with ``_via_{HTTP_METHOD}``.
    """
    method = api_def['method']
    family = api_def['family'].split(':')[1]
    path = _simplify_path(api_def['path'], family)
    known_params = {p['name'] for p in api_def['params']}
    namespaces, placeholders = _partition_path(path, known_params,
                                               api_def['slugs'])

    # I haven't seen more than 2.
    if placeholders:
        f_name = "_".join(namespaces)

        k = len(placeholders)
        if k == 1:
            f_name += "_by_" + placeholders[0]
        elif k == 2:
            b, a = placeholders
            f_name += "_by_{}_and_{}".format(a, b)
        else:
            raise ValueError("Unknown format")
    else:
        f_name = "_".join(namespaces)

    if f_name.startswith('_'):
        f_name = f_name[1:]

    if append_http_method:
        f_name += "_via_" + method.lower()

    if f_name == '':
        f_name = api_def['path']  # Singleton families?

    # XXX
    if not f_name.isidentifier():
        f_name = "".join(c for c in f_name if c.isidentifier())

    assert f_name.isidentifier(), f_name
    assert f_name, f_name

    return f_name


def _aggregate_defns(definitions):
    # parent -> child -> definition
    agg = defaultdict(lambda: defaultdict(list))

    # parent -> child -> name -> definition
    names = defaultdict(lambda: defaultdict(lambda: defaultdict(list)))

    # Aggregate names and definition collections.
    for parent_module, api_defs in definitions.items():
        parent_module = parent_module + "_api"
        for api_def in api_defs:
            sub_module = api_def['family'].split(":")[1]
            agg[parent_module][sub_module].append(api_def)

            name = _generate_func_name(api_def)
            names[parent_module][sub_module][name].append(api_def)

    return agg, names


def _annotate_nonunique_names(definitions):
    agg, names = _aggregate_defns(definitions)

    for parent_module, sub_module in names.items():
        for sub_module_name, functions in sub_module.items():
            for func_name, api_defs in functions.items():
                if len(api_defs) == 1:
                    continue  # Ready for source code generation

                methods = {d['method'] for d in api_defs}

                if len(methods) == len(api_defs):  # 1-to-1
                    if 'GET' in methods:
                        methods.remove('GET')

                    if len(methods) > 1 and 'DELETE' in methods:
                        methods.remove('DELETE')

                    assert len(methods) == 1

                    for api_def in api_defs:
                        if api_def['method'] not in methods:
                            api_def['_annotate_with_method'] = True
                else:
                    print(func_name, len(api_defs),
                          methods,
                          len(api_defs))
                    for api_def in api_defs:
                        print(api_def['reference_url'])
                    assert False, "UNKNOWN CONDITION"

    return agg



import pytest
import bz2
from unittest.mock import Mock, MagicMock, patch
import json
from brittle_wit_core import AppCredentials, ClientCredentials, TwitterRequest
from brittle_wit_core import TwitterError
from brittle_wit.streaming import (StreamReceiver,
                                   FeedSerializer,
                                   CallableStreamHandler,
                                   StreamProcessor,
                                   StreamingHTTPPipe,
                                   basic_streamer)
from tests.helpers import mock_stream, async_ignore_sleep, MockHTTPResp


def test_subscribe_raw():
    sp = StreamProcessor(None)

    sp.subscribe(CallableStreamHandler(lambda m: -1), as_json=False)

    assert not sp.requires_json


def test_subscribe_json():
    sp = StreamProcessor(None)

    sp.subscribe(CallableStreamHandler(lambda m: -1), as_json=True)

    assert sp.requires_json


def test_subscribe_receiver_must_implement_send():
    sp = StreamProcessor(None)
    with pytest.raises(AttributeError):
        sp.subscribe(None, True)


def test_no_duplicate_subscriptions():
    sp = StreamProcessor(None)
    handler = CallableStreamHandler(lambda m: -1)

    sp.subscribe(handler, True)
    with pytest.raises(RuntimeError):
        sp.subscribe(handler, True)


def test_unsubscribe_tracks_need_for_json():
    sp = StreamProcessor(None)

    json_handler = CallableStreamHandler(lambda m: -1)
    byte_handler = CallableStreamHandler(lambda m: -1)

    sp.subscribe(json_handler, True)
    sp.subscribe(byte_handler, False)

    assert sp.requires_json
    sp.unsubscribe(json_handler)
    assert not sp.requires_json


def test_no_unsubscribe_twice():
    sp = StreamProcessor(None)

    json_handler = CallableStreamHandler(lambda m: -1)
    byte_handler = CallableStreamHandler(lambda m: -1)

    sp.subscribe(json_handler, True)
    sp.unsubscribe(json_handler)

    with pytest.raises(RuntimeError):
        sp.unsubscribe(json_handler)


def test_receiving_multi():
    sp = StreamProcessor(None)

    byte_msgs, json_msgs = [], []
    json_handler = CallableStreamHandler(lambda m: json_msgs.append(m))
    byte_handler = CallableStreamHandler(lambda m: byte_msgs.append(m))
    sp.subscribe(byte_handler, False)
    sp.subscribe(json_handler, True)

    messages = [[1, 2, 3], [4, 5, 6]]
    for msg in messages:
        sp._send_to_all(json.dumps(msg).encode())

    assert json_msgs == messages
    assert byte_msgs, [json.dumps(m).encode() for m in messages]


@pytest.mark.asyncio
async def test_stream_to_subscribers():
    expected = [b"a", b"b"]

    sp = StreamProcessor(None)
    byte_msgs = []
    byte_handler = CallableStreamHandler(lambda m: byte_msgs.append(m))
    sp._twitter_stream = mock_stream(expected)
    sp.subscribe(byte_handler, False)
    await sp._stream_to_subscribers()
    assert byte_msgs == expected


@pytest.mark.asyncio
async def test_run_aborts_on_keyboard_interrupt():
    sp = StreamProcessor(None)
    sp._twitter_stream = mock_stream([KeyboardInterrupt()])
    with pytest.raises(KeyboardInterrupt):
        await sp.run()


@pytest.mark.asyncio
async def test_run_reraises_unguarded_exception():
    sp = StreamProcessor(None)
    sp._twitter_stream = mock_stream([AttributeError()])
    with pytest.raises(AttributeError):
        await sp.run()


async def _assert_run_expectations(expected, messages, **overrides):
    sp = StreamProcessor(mock_stream(messages))
    for k, v in overrides.items():
        setattr(sp, k, v)

    with async_ignore_sleep() as sleep_times:
        with patch.object(sp._twitter_stream, 'disconnect') as disconnect:
            with pytest.raises(RuntimeError) as e:
                await sp.run()
            assert "Sentinel" in str(e)

        assert len(disconnect.call_args_list) == len(expected)
        assert sleep_times() == expected


@pytest.mark.asyncio
async def test_run_on_timeout_error():
    import asyncio
    expected = [min(0.25 * i, 16.0) for i in range(1, 66)]
    messages = ([asyncio.TimeoutError()] * len(expected) +
                [RuntimeError("Sentinel")])
    await _assert_run_expectations(expected, messages)


@pytest.mark.asyncio
async def test_run_on_client_error():
    import aiohttp
    expected = [min(0.25 * i, 16.0) for i in range(1, 66)]
    messages = ([aiohttp.ClientError()] * len(expected) +
                [RuntimeError("Sentinel")])
    await _assert_run_expectations(expected, messages)


@pytest.mark.asyncio
async def test_run_on_420_error():
    err = TwitterError(None, None, MockHTTPResp(420, [], None), '')
    expected = [60, 120, 240, 480, 960, 960]
    messages = ([err] * len(expected) + [RuntimeError("Sentinel")])
    await _assert_run_expectations(expected, messages)


@pytest.mark.asyncio
async def test_run_on_other_twitter_error():
    err = TwitterError(None, None, MockHTTPResp(500, [], None), '')
    expected = [5, 10, 20, 40, 80, 160, 320, 320]
    messages = ([err] * len(expected) + [RuntimeError("Sentinel")])
    await _assert_run_expectations(expected, messages)


@pytest.mark.asyncio
async def test_run_on_rate_limit():
    err = TwitterError(None, None, MockHTTPResp(429, [], None), '')
    expected = [42] * 3
    messages = ([err] * len(expected) + [RuntimeError("Sentinel")])

    mock_limit = Mock()
    mock_limit.sleep_time_remaining = 42
    with patch('brittle_wit.rate_limit.RateLimit.from_response', return_value=mock_limit):
        await _assert_run_expectations(expected, messages, _resp=None)



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def direct_messages(*, since_id=_ELIDE, max_id=_ELIDE, count=_ELIDE,
                    include_entities=_ELIDE, skip_status=_ELIDE):
    """
    Returns the 20 most recent direct messages sent to the authenticating user.
    Includes detailed information about the sender and recipient user. You can
    request up to 200 direct messages per call, and only the most recent 200
    DMs will be available using this endpoint.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param count: Specifies the number of direct messages to try and retrieve,
        up to a maximum of 200. The value of count is best thought of as a
        limit to the number of Tweets to return because suspended or deleted
        content is removed after the count has been applied.

    :param include_entities: The entities node will not be included when set to
        false.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'since_id': since_id, 'max_id': max_id, 'count': count,
               'include_entities': include_entities, 'skip_status':
               skip_status}
    url = 'https://api.twitter.com/1.1/direct_messages.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages',
                           binding)


def events_list(*, count=_ELIDE, cursor=_ELIDE):
    """
    Returns all Direct Message events (both sent
    """
    binding = {'count': count, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/direct_messages/events/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-events-list',
                           binding)


def events_show(id):
    """
    Returns a single Direct Message event by the given id.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/direct_messages/events/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-events-show',
                           binding)


def sent(*, since_id=_ELIDE, max_id=_ELIDE, count=_ELIDE, page=_ELIDE,
         include_entities=_ELIDE):
    """
    Returns the 20 most recent direct messages sent by the authenticating user.
    Includes detailed information about the sender and recipient user. You can
    request up to 200 direct messages per call, up to a maximum of 200 outgoing
    DMs.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param count: Specifies the number of records to retrieve. Must be less
        than or equal to 200.

    :param page: Specifies the page of results to retrieve.

    :param include_entities: The entities node will not be included when set to
        false.
    """
    binding = {'since_id': since_id, 'max_id': max_id, 'count': count, 'page':
               page, 'include_entities': include_entities}
    url = 'https://api.twitter.com/1.1/direct_messages/sent.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-sent',
                           binding)


def show(id):
    """
    Returns a single direct message, specified by an id parameter. Like the

    :param id: The ID of the direct message.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/direct_messages/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-show',
                           binding)


def welcome_messages_list(*, count=_ELIDE, cursor=_ELIDE):
    """
    Returns a list of Welcome Messages.
    """
    binding = {'count': count, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-welcome-messages-list',
                           binding)


def welcome_messages_rules_list(*, count=_ELIDE, cursor=_ELIDE):
    """
    Returns a list of Welcome Message Rules.
    """
    binding = {'count': count, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-welcome-messages-rules-list',
                           binding)


def welcome_messages_rules_show(id):
    """
    Returns a Welcome Message Rule by the given id.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-welcome-messages-rules-show',
                           binding)


def welcome_messages_show(id):
    """
    Returns a Welcome Message by the given id.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:direct_messages',
                           'get-direct-messages-welcome-messages-show',
                           binding)


def destroy(id, *, include_entities=_ELIDE):
    """
    Destroys the direct message specified in the required ID parameter. The
    authenticating user must be the recipient of the specified direct message.

    :param id: The ID of the direct message to delete.

    :param include_entities: The entities node will not be included when set to
        false.
    """
    binding = {'id': id, 'include_entities': include_entities}
    url = 'https://api.twitter.com/1.1/direct_messages/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:direct_messages',
                           'post-direct-messages-destroy',
                           binding)


def events_newmessage_create():
    """
    Publishes a new
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/direct_messages/events/new.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:direct_messages',
                           'post-direct-messages-events-new-message-create',
                           binding)


def new(text, *, user_id=_ELIDE, screen_name=_ELIDE):
    """
    Sends a new direct message to the specified user from the authenticating
    user. Requires both the

    :param user_id: The ID of the user who should receive the direct message.
        Helpful for disambiguating when a valid user ID is also a valid screen
        name.

    :param screen_name: The screen name of the user who should receive the
        direct message. Helpful for disambiguating when a valid screen name is
        also a user ID.

    :param text: The text of your direct message. Be sure to URL encode as
        necessary, and keep the message under 140 characters.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'text': text}
    url = 'https://api.twitter.com/1.1/direct_messages/new.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:direct_messages',
                           'post-direct-messages-new',
                           binding)


def welcome_messages_new():
    """
    Creates a new Welcome Message that will be stored and sent in the future
    from the authenticating user in defined circumstances. Returns the message
    template in the requested format if successful. Supports publishing with
    the same elements as Direct Messages (e.g. Quick Replies, media
    attachments).
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/new.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:direct_messages',
                           'post-direct-messages-welcome-messages-new',
                           binding)


def welcome_messages_rules_new():
    """
    Creates a new Welcome Message Rule that determines which Welcome Message
    will be shown in a given conversation. Returns the created rule if
    successful.
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/new.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:direct_messages',
                           'post-direct-messages-welcome-messages-rules-new',
                           binding)


def welcome_messages_destroy(id):
    """
    Deletes a Welcome Message by the given id.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/destroy.json'
    return _TwitterRequest('DELETE',
                           url,
                           'rest:direct_messages',
                           'delete-direct-messages-welcome-messages-destroy',
                           binding)


def welcome_messages_rules_destroy(id):
    """
    Deletes a Welcome Message Rule by the given id.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/destroy.json'
    return _TwitterRequest('DELETE',
                           url,
                           'rest:direct_messages',
                           'delete-direct-messages-welcome-messages-rules-destroy',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/events/list.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/events/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/events/show.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/events/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/sent.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/sent'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/show.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/list.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/welcome_messages/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/list.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/welcome_messages/rules/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/show.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/welcome_messages/rules/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/show.json'] = 'https://dev.twitter.com/rest/reference/get/direct_messages/welcome_messages/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/direct_messages/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/events/new.json'] = 'https://dev.twitter.com/rest/reference/post/direct_messages/events/new'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/new.json'] = 'https://dev.twitter.com/rest/reference/post/direct_messages/new'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/new.json'] = 'https://dev.twitter.com/rest/reference/post/direct_messages/welcome_messages/new'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/new.json'] = 'https://dev.twitter.com/rest/reference/post/direct_messages/welcome_messages/rules/new'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/destroy.json'] = 'https://dev.twitter.com/rest/reference/del/direct_messages/welcome_messages/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/direct_messages/welcome_messages/rules/destroy.json'] = 'https://dev.twitter.com/rest/reference/del/direct_messages/welcome_messages/rules/destroy'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def home_timeline(*, count=_ELIDE, since_id=_ELIDE, max_id=_ELIDE,
                  trim_user=_ELIDE, exclude_replies=_ELIDE,
                  include_entities=_ELIDE):
    """
    Returns a collection of the most recent

    :param count: Specifies the number of records to retrieve. Must be less
        than or equal to 200. Defaults to 20. The value of count is best
        thought of as a limit to the number of tweets to return because
        suspended or deleted content is removed after the count has been
        applied.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param trim_user: When set to either true, t or 1, each Tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param exclude_replies: This parameter will prevent replies from appearing
        in the returned timeline. Using exclude_replies with the count
        parameter will mean you will receive up-to count Tweets — this is
        because the count parameter retrieves that many Tweets before filtering
        out retweets and replies.

    :param include_entities: The entities node will not be included when set to
        false.
    """
    binding = {'count': count, 'since_id': since_id, 'max_id': max_id,
               'trim_user': trim_user, 'exclude_replies': exclude_replies,
               'include_entities': include_entities}
    url = 'https://api.twitter.com/1.1/statuses/home_timeline.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-home-timeline',
                           binding)


def lookup(id, *, include_entities=_ELIDE, trim_user=_ELIDE, map=_ELIDE,
           include_ext_alt_text=_ELIDE):
    """
    Returns fully-hydrated

    :param id: A comma separated list of Tweet IDs, up to 100 are allowed in a
        single request.

    :param include_entities: The entities node that may appear within embedded
        statuses will not be included when set to false.

    :param trim_user: When set to either true, t or 1, each Tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param map: When using the map parameter, Tweets that do not exist or
        cannot be viewed by the current user will still have their key
        represented but with an explicitly null value paired with it

    :param include_ext_alt_text: If alt text has been added to any attached
        media entities, this parameter will return an ext_alt_text value in the
        top-level key for the media entity. If no value has been set, this will
        be returned as null
    """
    binding = {'id': id, 'include_entities': include_entities, 'trim_user':
               trim_user, 'map': map, 'include_ext_alt_text':
               include_ext_alt_text}
    url = 'https://api.twitter.com/1.1/statuses/lookup.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-lookup',
                           binding)


def mentions_timeline(*, count=_ELIDE, since_id=_ELIDE, max_id=_ELIDE,
                      trim_user=_ELIDE, include_entities=_ELIDE):
    """
    Returns the 20 most recent mentions (Tweets containing a users’s
    @screen_name) for the authenticating user.

    :param count: Specifies the number of Tweets to try and retrieve, up to a
        maximum of 200. The value of count is best thought of as a limit to the
        number of tweets to return because suspended or deleted content is
        removed after the count has been applied. We include retweets in the
        count, even if include_rts is not supplied. It is recommended you
        always send include_rts=1 when using this API method.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param trim_user: When set to either true, t or 1, each tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param include_entities: The entities node will not be included when set to
        false.
    """
    binding = {'count': count, 'since_id': since_id, 'max_id': max_id,
               'trim_user': trim_user, 'include_entities': include_entities}
    url = 'https://api.twitter.com/1.1/statuses/mentions_timeline.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-mentions-timeline',
                           binding)


def oembed(url, *, maxwidth=_ELIDE, hide_media=_ELIDE, hide_thread=_ELIDE,
           omit_script=_ELIDE, align=_ELIDE, related=_ELIDE, lang=_ELIDE,
           theme=_ELIDE, link_color=_ELIDE, widget_type=_ELIDE, dnt=_ELIDE):
    """
    Returns a single Tweet, specified by either a Tweet web URL or the Tweet
    ID, in an

    :param url: The URL of the Tweet to be embedded

    :param maxwidth: The maximum width of a rendered Tweet in whole pixels.
        This value must be between 220 and 550 inclusive. A supplied value
        under or over the allowed range will be returned as the minimum or
        maximum supported width respectively; the reset width value will be
        reflected in the returned width property. Note that Twitter does not
        support the oEmbed maxheight parameter. Tweets are fundamentally text,
        and are therefore of unpredictable height that cannot be scaled like an
        image or video. Relatedly, the oEmbed response will not provide a value
        for height. Implementations that need consistent heights for Tweets
        should refer to the hide_thread and hide_media parameters below

    :param hide_media: When set to true, t, or 1 links in a Tweet are not
        expanded to photo, video, or link previews

    :param hide_thread: When set to true, t, or 1 a collapsed version of the
        previous Tweet in a conversation thread will not be displayed when the
        requested Tweet is in reply to another Tweet

    :param omit_script: When set to true, t, or 1 the <script> responsible for
        loading widgets.js will not be returned. Your webpages should include
        their own reference to widgets.js for use across all Twitter widgets
        including Embedded Tweets

    :param align: Specifies whether the embedded Tweet should be floated left,
        right, or center in the page relative to the parent element. Valid
        values are left, right, center, and none

    :param related: A comma-separated list of Twitter usernames related to your
        content. This value will be forwarded to Tweet action intents if a
        viewer chooses to reply, like, or retweet the embedded Tweet

    :param lang: Request returned HTML and a rendered Tweet in the specified
        Twitter language supported by embedded Tweets

    :param theme: When set to dark, the Tweet is displayed with light text over
        a dark background

    :param link_color: Adjust the color of Tweet text links with a hexadecimal
        color value

    :param widget_type: Set to video to return a Twitter Video embed for the
        given Tweet

    :param dnt: When set to true, the Tweet and its embedded page on your site
        are not used for purposes that include personalized suggestions and
        personalized ads
    """
    binding = {'url': url, 'maxwidth': maxwidth, 'hide_media': hide_media,
               'hide_thread': hide_thread, 'omit_script': omit_script,
               'align': align, 'related': related, 'lang': lang, 'theme':
               theme, 'link_color': link_color, 'widget_type': widget_type,
               'dnt': dnt}
    url = 'https://publish.twitter.com/oembed'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-oembed',
                           binding)


def retweeters_ids(id, *, cursor=_ELIDE, stringify_ids=_ELIDE):
    """
    Returns a collection of up to 100 user IDs belonging to users who have
    retweeted the Tweet specified by the

    :param id: The numerical ID of the desired status.

    :param cursor: Causes the list of IDs to be broken into pages of no more
        than 100 IDs at a time. The number of IDs returned is not guaranteed to
        be 100 as suspended users are filtered out after connections are
        queried. If no cursor is provided, a value of -1 will be assumed, which
        is the first “page.” The response from the API will include a
        previous_cursor and next_cursor to allow paging back and forth. See our
        cursor docs for more information. While this method supports the cursor
        parameter, the entire result set can be returned in a single cursored
        collection. Using the count parameter with this method will not provide
        segmented cursors for use with this parameter.

    :param stringify_ids: Many programming environments will not consume Tweet
        ids due to their size. Provide this option to have ids returned as
        strings instead.
    """
    binding = {'id': id, 'cursor': cursor, 'stringify_ids': stringify_ids}
    url = 'https://api.twitter.com/1.1/statuses/retweeters/ids.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-retweeters-ids',
                           binding)


def retweets_by_id(id, *, count=_ELIDE, trim_user=_ELIDE):
    """
    Returns a collection of the 100 most recent retweets of the Tweet specified
    by the

    :param id: The numerical ID of the desired status.

    :param count: Specifies the number of records to retrieve. Must be less
        than or equal to 100.

    :param trim_user: When set to either true, t or 1, each tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    """
    binding = {'id': id, 'count': count, 'trim_user': trim_user}
    url = 'https://api.twitter.com/1.1/statuses/retweets/{id}.json'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-retweets-id',
                           binding)


def retweets_of_me(*, count=_ELIDE, since_id=_ELIDE, max_id=_ELIDE,
                   trim_user=_ELIDE, include_entities=_ELIDE,
                   include_user_entities=_ELIDE):
    """
    Returns the most recent Tweets authored by the authenticating user that
    have been retweeted by others. This timeline is a subset of the user’s

    :param count: Specifies the number of records to retrieve. Must be less
        than or equal to 100. If omitted, 20 will be assumed.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        which can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param trim_user: When set to either true, t or 1, each tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param include_entities: The tweet entities node will not be included when
        set to false.

    :param include_user_entities: The user entities node will not be included
        when set to false.
    """
    binding = {'count': count, 'since_id': since_id, 'max_id': max_id,
               'trim_user': trim_user, 'include_entities': include_entities,
               'include_user_entities': include_user_entities}
    url = 'https://api.twitter.com/1.1/statuses/retweets_of_me.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-retweets-of-me',
                           binding)


def show_by_id(id, *, trim_user=_ELIDE, include_my_retweet=_ELIDE,
               include_entities=_ELIDE, include_ext_alt_text=_ELIDE):
    """
    Returns a single

    :param id: The numerical ID of the desired Tweet.

    :param trim_user: When set to either true, t or 1, each Tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param include_my_retweet: When set to either true, t or 1, any Tweets
        returned that have been retweeted by the authenticating user will
        include an additional current_user_retweet node, containing the ID of
        the source status for the retweet.

    :param include_entities: The entities node will not be included when set to
        false.

    :param include_ext_alt_text: If alt text has been added to any attached
        media entities, this parameter will return an ext_alt_text value in the
        top-level key for the media entity. If no value has been set, this will
        be returned as null
    """
    binding = {'id': id, 'trim_user': trim_user, 'include_my_retweet':
               include_my_retweet, 'include_entities': include_entities,
               'include_ext_alt_text': include_ext_alt_text}
    url = 'https://api.twitter.com/1.1/statuses/show.json'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-show-id',
                           binding)


def user_timeline(*, user_id=_ELIDE, screen_name=_ELIDE, since_id=_ELIDE,
                  count=_ELIDE, max_id=_ELIDE, trim_user=_ELIDE,
                  exclude_replies=_ELIDE, include_rts=_ELIDE):
    """
    Returns a collection of the most recent

    :param user_id: The ID of the user for whom to return results.

    :param screen_name: The screen name of the user for whom to return results.

    :param since_id: Returns results with an ID greater than (that is, more
        recent than) the specified ID. There are limits to the number of Tweets
        that can be accessed through the API. If the limit of Tweets has
        occured since the since_id, the since_id will be forced to the oldest
        ID available.

    :param count: Specifies the number of Tweets to try and retrieve, up to a
        maximum of 200 per distinct request. The value of count is best thought
        of as a limit to the number of Tweets to return because suspended or
        deleted content is removed after the count has been applied. We include
        retweets in the count, even if include_rts is not supplied. It is
        recommended you always send include_rts=1 when using this API method.

    :param max_id: Returns results with an ID less than (that is, older than)
        or equal to the specified ID.

    :param trim_user: When set to either true, t or 1, each Tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param exclude_replies: This parameter will prevent replies from appearing
        in the returned timeline. Using exclude_replies with the count
        parameter will mean you will receive up-to count tweets — this is
        because the count parameter retrieves that many Tweets before filtering
        out retweets and replies.

    :param include_rts: When set to false, the timeline will strip any native
        retweets (though they will still count toward both the maximal length
        of the timeline and the slice selected by the count parameter). Note:
        If you’re using the trim_user parameter in conjunction with
        include_rts, the retweets will still contain a full user object.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'since_id':
               since_id, 'count': count, 'max_id': max_id, 'trim_user':
               trim_user, 'exclude_replies': exclude_replies, 'include_rts':
               include_rts}
    url = 'https://api.twitter.com/1.1/statuses/user_timeline.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:statuses',
                           'get-statuses-user-timeline',
                           binding)


def destroy_by_id(id, *, trim_user=_ELIDE):
    """
    Destroys the status specified by the required ID parameter. The
    authenticating user must be the author of the specified status. Returns the
    destroyed status if successful.

    :param id: The numerical ID of the desired status.

    :param trim_user: When set to either true, t or 1, each tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    """
    binding = {'id': id, 'trim_user': trim_user}
    url = 'https://api.twitter.com/1.1/statuses/destroy/{id}.json'
    url = url.format(**binding)
    return _TwitterRequest('POST',
                           url,
                           'rest:statuses',
                           'post-statuses-destroy-id',
                           binding)


def retweet_by_id(id, *, trim_user=_ELIDE):
    """
    Retweets a tweet. Returns the

    :param id: The numerical ID of the desired status.

    :param trim_user: When set to either true, t or 1, each tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    """
    binding = {'id': id, 'trim_user': trim_user}
    url = 'https://api.twitter.com/1.1/statuses/retweet/{id}.json'
    url = url.format(**binding)
    return _TwitterRequest('POST',
                           url,
                           'rest:statuses',
                           'post-statuses-retweet-id',
                           binding)


def unretweet_by_id(id, *, trim_user=_ELIDE):
    """
    Untweets a retweeted status. Returns the

    :param id: The numerical ID of the desired status.

    :param trim_user: When set to either true, t or 1, each tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.
    """
    binding = {'id': id, 'trim_user': trim_user}
    url = 'https://api.twitter.com/1.1/statuses/unretweet/{id}.json'
    url = url.format(**binding)
    return _TwitterRequest('POST',
                           url,
                           'rest:statuses',
                           'post-statuses-unretweet-id',
                           binding)


def update(status, *, in_reply_to_status_id=_ELIDE,
           possibly_sensitive=_ELIDE, lat=_ELIDE, long=_ELIDE,
           place_id=_ELIDE, display_coordinates=_ELIDE, trim_user=_ELIDE,
           media_ids=_ELIDE, enable_dm_commands=_ELIDE,
           fail_dm_commands=_ELIDE):
    """
    Updates the authenticating user’s current status, also known as Tweeting.

    :param status: The text of the status update, typically up to 140
        characters. URL encode as necessary. t.co link wrapping may affect
        character counts. There are some special commands in this field to be
        aware of. For instance, preceding a message with “D ” or “M ” and
        following it with a screen name can create a Direct Message to that
        user if the relationship allows for it. See the ‘enable_dm_commands’
        parameter for information on disabling Direct Message creation to allow
        any text to be created as a Tweet.

    :param in_reply_to_status_id: The ID of an existing status that the update
        is in reply to. Note: This parameter will be ignored unless the author
        of the Tweet this parameter references is mentioned within the status
        text. Therefore, you must include @username, where username is the
        author of the referenced Tweet, within the update.

    :param possibly_sensitive: If you upload Tweet media that might be
        considered sensitive content such as nudity, violence, or medical
        procedures, you should set this value to true. See Media setting and
        best practices for more context.

    :param lat: The latitude of the location this Tweet refers to. This
        parameter will be ignored unless it is inside the range -90.0 to +90.0
        (North is positive) inclusive. It will also be ignored if there isn’t a
        corresponding long parameter.

    :param long: The longitude of the location this Tweet refers to. The valid
        ranges for longitude is -180.0 to +180.0 (East is positive) inclusive.
        This parameter will be ignored if outside that range, if it is not a
        number, if geo_enabled is disabled, or if there not a corresponding lat
        parameter.

    :param place_id: A place in the world.

    :param display_coordinates: Whether or not to put a pin on the exact
        coordinates a Tweet has been sent from.

    :param trim_user: When set to either true, t or 1, each Tweet returned in a
        timeline will include a user object including only the status authors
        numerical ID. Omit this parameter to receive the complete user object.

    :param media_ids: A list of media_ids to associate with the Tweet. You may
        include up to 4 photos or 1 animated GIF or 1 video in a Tweet. See
        Uploading Media for further details on uploading media.

    :param enable_dm_commands: When set to true, enables shortcode commands for
        sending Direct Messages as part of the status text to send a Direct
        Message to a user. When set to false, disables this behavior and
        includes any leading characters in the status text that is posted

    :param fail_dm_commands: When set to true, causes any status text that
        starts with shortcode commands to return an API error. When set to
        false, allows shortcode commands to be sent in the status text and
        acted on by the API.
    """
    binding = {'status': status, 'in_reply_to_status_id':
               in_reply_to_status_id, 'possibly_sensitive':
               possibly_sensitive, 'lat': lat, 'long': long, 'place_id':
               place_id, 'display_coordinates': display_coordinates,
               'trim_user': trim_user, 'media_ids': media_ids,
               'enable_dm_commands': enable_dm_commands, 'fail_dm_commands':
               fail_dm_commands}
    url = 'https://api.twitter.com/1.1/statuses/update.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:statuses',
                           'post-statuses-update',
                           binding)


def update_with_mediadeprecated(status, media, *, possibly_sensitive=_ELIDE,
                                in_reply_to_status_id=_ELIDE, lat=_ELIDE,
                                long=_ELIDE, place_id=_ELIDE,
                                display_coordinates=_ELIDE,
                                enable_dm_commands=_ELIDE,
                                fail_dm_commands=_ELIDE):
    """


    :param status: The text of your status update. URL encode as necessary.
        t.co link wrapping may affect character counts if the post contains
        URLs. You must additionally account for the
        characters_reserved_per_media per uploaded media, additionally
        accounting for space characters in between finalized URLs. Note :
        Request the GET help / configuration endpoint to get the current
        characters_reserved_per_media and max_media_per_upload values.

    :param media: Up to max_media_per_upload files may be specified in the
        request, each named media[]. Supported image formats are PNG, JPG and
        GIF, including animated GIFs of up to 3MB. This data must be either the
        raw image bytes or encoded as base64. Note : Request the GET help /
        configuration endpoint to get the current max_media_per_upload and
        photo_size_limit values.

    :param possibly_sensitive: Set to true for content which may not be
        suitable for every audience.

    :param in_reply_to_status_id: The ID of an existing status that the update
        is in reply to. Note : This parameter will be ignored unless the author
        of the Tweet this parameter references is mentioned within the status
        text. Therefore, you must include @username, where username is the
        author of the referenced Tweet, within the update.

    :param lat: The latitude of the location this Tweet refers to. This
        parameter will be ignored unless it is inside the range -90.0 to +90.0
        (North is positive) inclusive. It will also be ignored if there isn’t a
        corresponding long parameter.

    :param long: The longitude of the location this Tweet refers to. The valid
        ranges for longitude is -180.0 to +180.0 (East is positive) inclusive.
        This parameter will be ignored if outside that range, not a number,
        geo_enabled is disabled, or if there not a corresponding lat parameter.

    :param place_id: A place in the world identified by a Twitter place ID.
        Place IDs can be retrieved from geo/reverse_geocode.

    :param display_coordinates: Whether or not to put a pin on the exact
        coordinates a Tweet has been sent from.

    :param enable_dm_commands: When set to true, enables shortcode commands for
        sending Direct Messages as part of the status text to send a Direct
        Message to a user. When set to false, disables this behavior and
        includes any leading characters in the status text.

    :param fail_dm_commands: When set to true, causes any status text that
        starts with shortcode commands to return an API error. When set to
        false, allows shortcode commands to be sent in the status text and
        acted on by the API.
    """
    binding = {'status': status, 'media[]': media, 'possibly_sensitive':
               possibly_sensitive, 'in_reply_to_status_id':
               in_reply_to_status_id, 'lat': lat, 'long': long, 'place_id':
               place_id, 'display_coordinates': display_coordinates,
               'enable_dm_commands': enable_dm_commands, 'fail_dm_commands':
               fail_dm_commands}
    url = 'https://api.twitter.com/1.1/statuses/update_with_media.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:statuses',
                           'post-statuses-update-with-media-deprecated',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/home_timeline.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/home_timeline'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/lookup.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/lookup'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/mentions_timeline.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/mentions_timeline'
_TwitterRequest.DOC_URLS['https://publish.twitter.com/oembed'] = 'https://dev.twitter.com/rest/reference/get/statuses/oembed'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/retweeters/ids.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/retweeters/ids'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/retweets/{id}.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/retweets/id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/retweets_of_me.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/retweets_of_me'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/show.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/show/id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/user_timeline.json'] = 'https://dev.twitter.com/rest/reference/get/statuses/user_timeline'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/destroy/{id}.json'] = 'https://dev.twitter.com/rest/reference/post/statuses/destroy/id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/retweet/{id}.json'] = 'https://dev.twitter.com/rest/reference/post/statuses/retweet/id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/unretweet/{id}.json'] = 'https://dev.twitter.com/rest/reference/post/statuses/unretweet/id'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/update.json'] = 'https://dev.twitter.com/rest/reference/post/statuses/update'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/statuses/update_with_media.json'] = 'https://dev.twitter.com/rest/reference/post/statuses/update_with_media'



#!/usr/bin/env python

import json
import os
from collections import defaultdict
from api_code_gen import (generate_source, generate_modules,
                          _annotate_nonunique_names)


SELF_DIR = os.path.dirname(os.path.realpath(__file__))
LIB_BASE = os.path.join(SELF_DIR, os.pardir, "brittle_wit")
CLEAN_DIR = os.path.join(SELF_DIR, "data", "clean")
API_PATH = os.path.join(CLEAN_DIR, "api.json")


# These endpoints prefixed with 1 seem to be identical to those
# that are not prefixed with 1. I think it's legacy. I need to look
# into it more, but for now, I'll skip the edge case in src generation.
IGNORE_REF_URLS = {"https://ads-api.twitter.com/1/accounts/:account_id/tailored_audiences/:id",
                   "https://dev.twitter.com/ads/reference/1/get/accounts/account_id/tailored_audiences/id",
                   "https://dev.twitter.com/ads/reference/1/del/accounts/account_id/tailored_audiences/id",
                   "https://dev.twitter.com/ads/reference/1/get/accounts/account_id/tailored_audiences",
                   "https://dev.twitter.com/ads/reference/1/post/accounts/account_id/tailored_audiences"}


def correcting_code_gen(api_def):
    return generate_source(api_def, '_annotate_nonunique_names' in api_def)


if __name__ == '__main__':
    with open(API_PATH) as fp:
        definitions = {}
        for group, api_defns in json.load(fp).items():
            valid_api_defs = []
            for api_def in api_defns:
                if api_def['reference_url'] not in IGNORE_REF_URLS:
                    valid_api_defs.append(api_def)
            definitions[group] = valid_api_defs

    _annotate_nonunique_names(definitions)
    generate_modules(LIB_BASE, definitions)



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def ids(*, stringify_ids=_ELIDE, cursor=_ELIDE):
    """
    Returns an array of numeric user ids the authenticating user is blocking.

    :param stringify_ids: Many programming environments will not consume our
        ids due to their size. Provide this option to have ids returned as
        strings instead. Read more about Twitter IDs.

    :param cursor: Causes the list of IDs to be broken into pages of no more
        than 5000 IDs at a time. The number of IDs returned is not guaranteed
        to be 5000 as suspended users are filtered out after connections are
        queried. If no cursor is provided, a value of -1 will be assumed, which
        is the first “page.” The response from the API will include a
        previous_cursor and next_cursor to allow paging back and forth. See
        Using cursors to navigate collections for more information.
    """
    binding = {'stringify_ids': stringify_ids, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/blocks/ids.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:blocks',
                           'get-blocks-ids',
                           binding)


def list(*, include_entities=_ELIDE, skip_status=_ELIDE, cursor=_ELIDE):
    """
    Returns a collection of

    :param include_entities: The entities node will not be included when set to
        false.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.

    :param cursor: Causes the list of blocked users to be broken into pages of
        no more than 5000 IDs at a time. The number of IDs returned is not
        guaranteed to be 5000 as suspended users are filtered out after
        connections are queried. If no cursor is provided, a value of -1 will
        be assumed, which is the first “page.” The response from the API will
        include a previous_cursor and next_cursor to allow paging back and
        forth. See Using cursors to navigate collections for more information.
    """
    binding = {'include_entities': include_entities, 'skip_status':
               skip_status, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/blocks/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:blocks',
                           'get-blocks-list',
                           binding)


def create(*, screen_name=_ELIDE, user_id=_ELIDE, include_entities=_ELIDE,
           skip_status=_ELIDE):
    """
    Blocks the specified user from following the authenticating user. In
    addition the blocked user will not show in the authenticating users
    mentions or timeline (unless retweeted by another user). If a follow or
    friend relationship exists it is destroyed.

    :param screen_name: The screen name of the potentially blocked user.
        Helpful for disambiguating when a valid screen name is also a user ID.

    :param user_id: The ID of the potentially blocked user. Helpful for
        disambiguating when a valid user ID is also a valid screen name.

    :param include_entities: The entities node will not be included when set to
        false.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id,
               'include_entities': include_entities, 'skip_status':
               skip_status}
    url = 'https://api.twitter.com/1.1/blocks/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:blocks',
                           'post-blocks-create',
                           binding)


def destroy(*, screen_name=_ELIDE, user_id=_ELIDE, include_entities=_ELIDE,
            skip_status=_ELIDE):
    """
    Un-blocks the user specified in the ID parameter for the authenticating
    user. Returns the un-blocked user in the requested format when successful.
    If relationships existed before the block was instated, they will not be
    restored.

    :param screen_name: The screen name of the potentially blocked user.
        Helpful for disambiguating when a valid screen name is also a user ID.

    :param user_id: The ID of the potentially blocked user. Helpful for
        disambiguating when a valid user ID is also a valid screen name.

    :param include_entities: The entities node will not be included when set to
        false.

    :param skip_status: When set to either true, t or 1 statuses will not be
        included in the returned user objects.
    """
    binding = {'screen_name': screen_name, 'user_id': user_id,
               'include_entities': include_entities, 'skip_status':
               skip_status}
    url = 'https://api.twitter.com/1.1/blocks/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:blocks',
                           'post-blocks-destroy',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/blocks/ids.json'] = 'https://dev.twitter.com/rest/reference/get/blocks/ids'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/blocks/list.json'] = 'https://dev.twitter.com/rest/reference/get/blocks/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/blocks/create.json'] = 'https://dev.twitter.com/rest/reference/post/blocks/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/blocks/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/blocks/destroy'



#!/usr/bin/env python
###############################################################################
# This script fetches the html docs for all endpoints in the twitter api.
###############################################################################
import json
import os
import requests
from lxml.html import fromstring


SELF_DIR = os.path.dirname(os.path.realpath(__file__))
RAW_DIR = os.path.join(SELF_DIR, "data", "raw")

ADS_DOC_PATHS = ["https://dev.twitter.com/ads/reference/get",
                 "https://dev.twitter.com/ads/reference/post"
                 "https://dev.twitter.com/ads/reference/del"
                 "https://dev.twitter.com/ads/reference/put"]

IGNORE_PATHS = set(ADS_DOC_PATHS)

REFERENCE_ROOTS = {"rest": ["https://dev.twitter.com/rest/reference"],
                   "streaming": ["https://dev.twitter.com/streaming/reference"],
                   "webhooks": ["https://dev.twitter.com/webhooks/reference"],
                   "ads": ADS_DOC_PATHS}


def fetch_endpoints_from(root_url):
    doc = fromstring(requests.get(root_url).content)
    doc.make_links_absolute(root_url)

    valid_methods = {'DELETE', 'GET', 'POST', 'PUT'}

    endpoints = []
    for a in doc.xpath("//a"):
        txt, href = a.text, a.attrib['href']

        if href in IGNORE_PATHS:
            continue

        if '/reference/' in href and a.text.split(" ")[0] in valid_methods:
            method, *service = a.text.split(" ")
            endpoints.append({'method': method,
                              'service': " ".join(service),
                              'reference_url': href})

    return endpoints


def fetch_and_save_all(fp, group, reference_root_urls):
    already_fetched = set()

    endpoints = []
    for reference_root_url in reference_root_urls:
        endpoints.extend(fetch_endpoints_from(reference_root_url))

    for d in endpoints:
        print("...", d['reference_url'])
        resp = requests.get(d['reference_url'])
        d['_raw_html'] = resp.content.decode(resp.encoding)
        d['group'] = group
        if d['reference_url'] not in already_fetched:
            fp.write(json.dumps(d) + "\n")
            already_fetched.add(d['reference_url'])


if __name__ == '__main__':
    os.makedirs(RAW_DIR, exist_ok=True)

    with open(os.path.join(RAW_DIR, "raw_html.jsonl"), "w") as fp:
        for group, reference_root_urls in REFERENCE_ROOTS.items():
            print("Fetching for", group)
            fetch_and_save_all(fp, group, reference_root_urls)
            print("")



import json
import pytest
import os
from brittle_wit.streaming.receivers import (StreamReceiver,
                                             flexible_open,
                                             FeedSerializer,
                                             PartitioningFeedSerializer,
                                             CallableStreamHandler)


def test_stream_receiver_send_does_nothing_for_base():
    StreamReceiver().send(b"asdf")


def test_stream_receiver_close_does_nothing_for_base():
    StreamReceiver().close()


@pytest.mark.parametrize('extension', ['bz2', 'gz', 'xz', 'txt'])
def test_flexible_open(tmpdir, extension):
    output_path = str(tmpdir.join("hello." + extension))

    # Open.
    with flexible_open(output_path) as fp:
        fp.write(b"hello world")
    size = os.path.getsize(output_path)
    assert size > 0

    # Append.
    with flexible_open(output_path) as fp:
        fp.write(b"hello world")
    assert os.path.getsize(output_path) > size

    # Overwrite.
    with flexible_open(output_path, append=False) as fp:
        fp.write(b"hello")
    assert os.path.getsize(output_path) < size


def test_feed_serializer(tmpdir):
    output_path = str(tmpdir.join("feed.jsonl"))
    messages = [[1, 2], [3, 4]]

    feed_serializer = FeedSerializer(output_path)

    for msg in messages:
        feed_serializer.send(msg)

    feed_serializer.close()

    with open(output_path, 'rb') as fp:
        res = [json.loads(s.decode('utf8')) for s in fp.readlines()]
        assert res == messages


def test_partitioning_feed_serializer(tmpdir):
    path = tmpdir.join("{}.jsonl")
    serializer = PartitioningFeedSerializer(5, str(path))
    for i in range(11):
        serializer.send({"i": i})

    files = tmpdir.listdir()
    assert len(files) == 3


def test_callable_stream_handler():
    received = []
    rcvr = CallableStreamHandler(lambda msg: received.append(msg))

    messages = [[1, 2], [3, 4]]
    for msg in messages:
        rcvr.send(msg)

    assert received == messages



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def entries(id, *, count=_ELIDE, max_position=_ELIDE, min_position=_ELIDE):
    """
    Retrieve the identified Collection, presented as a list of the Tweets
    curated within.

    :param id: The identifier of the Collection to return results for.

    :param count: Specifies the maximum number of results to include in the
        response. Specify a count between 1 and 200. A next_cursor value will
        be provided in the response if additional results are available.

    :param max_position: Returns results with a position value less than or
        equal to the specified position.

    :param min_position: Returns results with a position greater than the
        specified position.
    """
    binding = {'id': id, 'count': count, 'max_position': max_position,
               'min_position': min_position}
    url = 'https://api.twitter.com/1.1/collections/entries.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:collections',
                           'get-collections-entries',
                           binding)


def list(user_id, screen_name, *, tweet_id=_ELIDE, count=_ELIDE,
         cursor=_ELIDE):
    """
    Find Collections created by a specific user or containing a specific
    curated Tweet.

    :param user_id: The ID of the user for whom to return results for.

    :param screen_name: The screen name of the user for whom to return results
        for.

    :param tweet_id: The identifier of the Tweet for which to return results
        for.

    :param count: Specifies the maximum number of results to include in the
        response. Specify a count between 1 and 200. A next_cursor value will
        be provided in the response if additional results are available.

    :param cursor: A string identifying the segment of the current result set
        to retrieve. Values for this parameter are yielded in the cursors node
        attached to response objects. Usage of the count parameter affects
        cursoring.
    """
    binding = {'user_id': user_id, 'screen_name': screen_name, 'tweet_id':
               tweet_id, 'count': count, 'cursor': cursor}
    url = 'https://api.twitter.com/1.1/collections/list.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:collections',
                           'get-collections-list',
                           binding)


def show(id):
    """
    Retrieve information associated with a specific Collection.

    :param id: The identifier of the Collection to return results for.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/collections/show.json'
    return _TwitterRequest('GET',
                           url,
                           'rest:collections',
                           'get-collections-show',
                           binding)


def create(name, *, description=_ELIDE, url=_ELIDE, timeline_order=_ELIDE):
    """
    Create a Collection owned by the currently authenticated user.

    :param name: The title of the collection being created, in 25 characters or
        less.

    :param description: A brief description of this collection in 160
        characters or fewer.

    :param url: A fully-qualified URL to associate with this collection.

    :param timeline_order: Order Tweets chronologically or in the order they
        are added to a Collection. curation_reverse_chron - order added
        (default) tweet_chron - oldest first tweet_reverse_chron - most recent
        first
    """
    binding = {'name': name, 'description': description, 'url': url,
               'timeline_order': timeline_order}
    url = 'https://api.twitter.com/1.1/collections/create.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-create',
                           binding)


def destroy(id):
    """
    Permanently delete a Collection owned by the currently authenticated user.

    :param id: The identifier of the Collection to destroy.
    """
    binding = {'id': id}
    url = 'https://api.twitter.com/1.1/collections/destroy.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-destroy',
                           binding)


def entries_add(id, tweet_id, *, relative_to=_ELIDE, above=_ELIDE):
    """
    Add a specified Tweet to a Collection.

    :param id: The identifier of the Collection receiving the Tweet.

    :param tweet_id: The identifier of the Tweet to add to the Collection.

    :param relative_to: The identifier of the Tweet used for relative
        positioning in a curation_reverse_chron ordered collection.

    :param above: Set to false to insert the specified tweet_id below the
        relative_to Tweet in the collection. Default: true
    """
    binding = {'id': id, 'tweet_id': tweet_id, 'relative_to': relative_to,
               'above': above}
    url = 'https://api.twitter.com/1.1/collections/entries/add.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-entries-add',
                           binding)


def entries_curate():
    """
    Curate a Collection by adding or removing Tweets in bulk. Updates must be
    limited to 100 cumulative additions or removals per request.
    """
    binding = {}
    url = 'https://api.twitter.com/1.1/collections/entries/curate.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-entries-curate',
                           binding)


def entries_move(id, tweet_id, relative_to, *, above=_ELIDE):
    """
    Move a specified Tweet to a new position in a

    :param id: The identifier of the Collection receiving the Tweet.

    :param tweet_id: The identifier of the Tweet to add to the Collection.

    :param relative_to: The identifier of the Tweet used for relative
        positioning.

    :param above: Set to false to insert the specified tweet_id below the
        relative_to Tweet in the collection. Default: true
    """
    binding = {'id': id, 'tweet_id': tweet_id, 'relative_to': relative_to,
               'above': above}
    url = 'https://api.twitter.com/1.1/collections/entries/move.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-entries-move',
                           binding)


def entries_remove(id, tweet_id):
    """
    Remove the specified Tweet from a Collection.

    :param id: The identifier of the target Collection.

    :param tweet_id: The identifier of the Tweet to remove.
    """
    binding = {'id': id, 'tweet_id': tweet_id}
    url = 'https://api.twitter.com/1.1/collections/entries/remove.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-entries-remove',
                           binding)


def update(id, *, name=_ELIDE, description=_ELIDE, url=_ELIDE):
    """
    Update information concerning a Collection owned by the currently
    authenticated user.

    :param id: The identifier of the Collection to modify.

    :param name: The title of the Collection being created, in 25 characters or
        fewer.

    :param description: A brief description of this Collection in 160
        characters or fewer.

    :param url: A fully-qualified URL to associate with this Collection.
    """
    binding = {'id': id, 'name': name, 'description': description, 'url': url}
    url = 'https://api.twitter.com/1.1/collections/update.json'
    return _TwitterRequest('POST',
                           url,
                           'rest:collections',
                           'post-collections-update',
                           binding)


_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/entries.json'] = 'https://dev.twitter.com/rest/reference/get/collections/entries'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/list.json'] = 'https://dev.twitter.com/rest/reference/get/collections/list'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/show.json'] = 'https://dev.twitter.com/rest/reference/get/collections/show'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/create.json'] = 'https://dev.twitter.com/rest/reference/post/collections/create'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/destroy.json'] = 'https://dev.twitter.com/rest/reference/post/collections/destroy'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/entries/add.json'] = 'https://dev.twitter.com/rest/reference/post/collections/entries/add'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/entries/curate.json'] = 'https://dev.twitter.com/rest/reference/post/collections/entries/curate'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/entries/move.json'] = 'https://dev.twitter.com/rest/reference/post/collections/entries/move'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/entries/remove.json'] = 'https://dev.twitter.com/rest/reference/post/collections/entries/remove'
_TwitterRequest.DOC_URLS['https://api.twitter.com/1.1/collections/update.json'] = 'https://dev.twitter.com/rest/reference/post/collections/update'



import pytest
import aiohttp
import asyncio
from unittest.mock import patch, MagicMock
from brittle_wit import AppCredentials, ClientCredentials
from brittle_wit.app import App, ClientContext, RUNNING, PRESTINE, FINISHED
from brittle_wit.executors import ManagedClientRequestProcessors
from tests.helpers import AsyncSafeTestCase, async_patch


def test_run_no_tasks(app):
    asyncio.set_event_loop(asyncio.new_event_loop())
    app.run()
    assert app._state == FINISHED


def test_run_maintenance_gets_canceled(app):
    asyncio.set_event_loop(asyncio.new_event_loop())

    res = []

    async def maintenance():
        try:
            await asyncio.sleep(10)
        except asyncio.CancelledError:
            res.append(42)

    app.schedule_maintenance(maintenance(), True)
    app.run()
    assert app._state == FINISHED


def test_setup_and_run(app_cred):
    asyncio.set_event_loop(asyncio.new_event_loop())
    res = []

    async def f():
        res.append(42)

    with App.with_defaults(app_cred).setup_and_run() as app:
        app.schedule(f())

    assert res == [42]



import aiohttp
from brittle_wit_core import TwitterRequest, AppCredentials, ClientCredentials
from brittle_wit.executors import twitter_req_to_http_req
from test.helpers import *


app_cred = AppCredentials("app", "secret")
client_cred = ClientCredentials(1, "client", "secret")


# -----------------------------------------------------------------------------
# This are weak tests but they should help prevent trivial regressions.
# -----------------------------------------------------------------------------


def test_twitter_req_to_http_req_get(capsys):
    tr = TwitterRequest("GET", "http://url.com/", "service", "family")
    with aiohttp.ClientSession() as session:
        req = twitter_req_to_http_req(session, app_cred, client_cred, tr)
        assert isinstance(req, aiohttp.client._RequestContextManager)


def test_twitter_req_to_http_req_post():
    tr = TwitterRequest("POST", "http://url.com/", "service", "family")
    with aiohttp.ClientSession() as session:
        req = twitter_req_to_http_req(session, app_cred, client_cred, tr)
        assert isinstance(req, aiohttp.client._RequestContextManager)



import asyncio
import aiohttp
import async_timeout
import time
import logging


from brittle_wit_core import (generate_req_headers,
                              TwitterResponse,
                              TwitterError,
                              BrittleWitError,
                              WrappedException)
from brittle_wit.helpers import retrying, CircularQueue
from brittle_wit.rate_limit import RateLimit, FIFTEEN_MINUTES
from lockchain import ChainManager


LOGGER = logging.getLogger('brittle_wit')

ANY_CREDENTIALS = None


def twitter_req_to_http_req(session, app_cred, client_cred, twitter_req, **overrides):
    """
    Convert a TwitterRequest into a aiohttp request.

    :param session: an aiohttp.ClientSession object
    :param app_cred: an `AppCredentials` object
    :param client_cred: an `ClientCredentials` object
    :param twitter_req: an `TwitterRequest` object
    :param overrides: override keys for the request

    :return: the aiohttp request
    """
    headers = generate_req_headers(twitter_req, app_cred, client_cred)
    payload_k = 'data' if twitter_req.method == 'POST' else 'params'
    kwargs = {payload_k: twitter_req.params, 'headers': headers}
    kwargs.update(overrides)

    return session.request(twitter_req.method, twitter_req.url, **kwargs)


async def execute_req(client_cred, twitter_req, http_req, rate_limit):
    """
    Execute a request with no exception handling.

    :param client_cred: the client_cred used to build the http_req
    :param twitter_req: the original TwitterRequest
    :param http_req: the HttpRequest with encoded authentication information
    :param rate_limit: the RateLimit object associated with the endpoint and
        client_cred
    :return: a TwitterResponse
    """

    # Send request over the wire.
    async with http_req as resp:

        # Update rate limits ASAP. Note: The prior call to comply
        # decremented the rate limits already. If the context manager
        # threw an exception on entry, it is still in compliance.
        rate_limit.update(resp)

        # An error occurred.
        if resp.status != 200:
            raise TwitterError(client_cred, twitter_req, resp,
                               await resp.text())

        # The status code indicates a good response.
        # Choose appropriate parsing strategy.
        if twitter_req.parse_as == 'json':
            body = await resp.json()
        else:
            body = await resp.text()

        return TwitterResponse(twitter_req, resp, body)


class ClientRequestProcessor:
    """
    The ClientRequestProcessor manages rate limits while executing requests.

    When the caller issues a request for execution, the request processor
    enforces rate limitations. Such enforcement occurs in the context of the
    given credentials and for a given service. If multiple callers request
    execution at the same time (for a given set of credentials and service),
    the requests execute FIFO.

    There should only be one ClientRequestProcessor per set of credentials!

    Periodically call ``#cleanup()``. When `#is_removable()` is true, there
    is no useful state information in the processor, and you can let Python
    garbage collect the object. This allows for stochastic reclaimation of
    memory.
    """

    def __init__(self, client_cred):
        """
        :param client_cred: the ClientCredentials to manage
        """
        self._client_cred = client_cred
        self._chain_manager = ChainManager()
        self._rate_limits = {}

    @property
    def user_id(self):
        return self._client_cred.user_id

    def is_removable(self):
        """
        :return: True if this executor is not in use and is monitoring no
            rate limits
        """
        return self.is_not_processing() and not self._rate_limits

    def is_not_processing(self):
        """
        :return: True if no caller is using this processors' credentials for
            any endpoint otherwise True
        """
        return self._chain_manager.all_chains_empty()

    def is_immediately_available_for(self, service):
        """
        :param service: the API service name, as per the Twitter docs
        :return: True if no caller is using this objects associated credentials
            for the given service AND that service is not rate limited
        """
        if self._chain_manager.is_chain_empty(service):
            if service not in self._rate_limits:
                return True
            else:
                return not self._rate_limits[service].is_exhausted
        return False

    def _rate_limit_for(self, service):
        """
        Get existing or create new RateLimit for the given service.

        Only this object should manage the RateLimits, hence a private method.

        :param service: API endpoint service
        :return: the RateLimit object associated with the service
        """
        limit = self._rate_limits.get(service)

        if limit is not None:
            return limit
        else:
            max_requests = RateLimit.CLIENT_LIMITS[service]
            limit = RateLimit(max_requests,
                              max_requests,
                              time.time() + FIFTEEN_MINUTES)
            self._rate_limits[service] = limit
            return limit

    def cleanup(self, now=None):
        """
        Remove all rate limits which have expired.

        :param now: the current time as per time.time()
        """
        if not self._rate_limits:
            return

        now = now or time.time()

        remove_ks = []
        for service, rate_limit in self._rate_limits.items():
            # If the reset time is in the past, remove it.
            if rate_limit.reset_time < now:
                remove_ks.append(service)

        for k in remove_ks:
            del self._rate_limits[k]

    async def _execute(self, session, app_cred, twitter_req, timeout=None):
        """
        :param session: an aiohttp.ClientSession object
        :param app_cred: an `AppCredentials` object
        :param twitter_req: an `TwitterRequest` object
        :param timeout: maximium time to wait for a response EVEN IF TWITTER
            IS STILL SENDING INFORMATION. (default None for no timeout)
        """
        service = twitter_req.service

        # This thows after some amount of time, even if it's still reading!
        async with async_timeout.timeout(timeout):

            async with self._chain_manager.chain_on(service):
                # The caller is now in charge of this service.
                rate_limit = self._rate_limit_for(service)

                # Block request execution until not rate limited.
                await rate_limit.comply()

                http_req = twitter_req_to_http_req(session,
                                                   app_cred,
                                                   self._client_cred,
                                                   twitter_req)

                # This coroutine will raise an exception for non-200 statuses.
                return await execute_req(self._client_cred,
                                         twitter_req,
                                         http_req,
                                         rate_limit)

    async def execute(self, session, app_cred, twitter_req, timeout=120,
                      n_retries=5, sleep_time=60):
        def bound_execute():
            return self._execute(session, app_cred, twitter_req, timeout)
        return await retrying(bound_execute, n_retries, sleep_time=sleep_time)


class ManagedClientRequestProcessors:
    """
    Manages client request processors

    Any Credential Management
    -------------------------

    For many tasks, you don't need specific credentials, you just need some
    credentials. For example, if you and your social network analyst friends
    wanted to analyze the followers of Tech Crunch, you need to collect all
    the user objects. Given that there are millions of them that could take
    time. But, if you already have the user id's you can partition the
    collection and run it in parallel. The specific credentials don't matter.
    (Well, ignoring the case of a follower blocking that person's access.)

    Finding an unused set of credentials is non-trivial. Managing a ready-queue
    for each service is expensive in terms of space when there are many users.
    As an alternative, I rely on probability. (That is, until I think
    more and find a better solution.)

    The any_cred_queue maintains a circular buffer for each set of
    credentials. When a client asks for AnyCredentials, the queue pops
    credentials off the top. When it asks again, it pushes that initial item
    to the end of the queue, and pops the head again.

    This is far from perfect. For example, this could select a token that is
    currently rate-limited even though non-limited ones remain. But, it's
    cheap, and does well, probabilistically. Additionally, I believe this
    should perform better as the number of managed users rises. (I'll think
    more about it; then, pen an arXiv article.)
    """

    def __init__(self):
        self._credentials, self._any_cred_queue = {}, CircularQueue()

    def is_managed(self, cred):
        return cred in self._credentials

    def add(self, cred, any_cred_eligible=True):
        """
        Create and manage a request processor for the given credentials.

        If you add already added credentials, it raises a runtime exception.

        :param cred: the client's credentials
        :param any_cred_eligible: if True, then an app may use this
            client's credentials promiscuously; otherwise, only the
            client can use it.
        """
        # XXX: TODO: What? Why? I'm sure I did this for a reason but...
        if self.is_managed(cred):
            raise RuntimeError("Credential already managed.")

        client_req_processor = ClientRequestProcessor(cred)
        self._credentials[cred] = client_req_processor

        if any_cred_eligible:
            self._any_cred_queue.add(cred)

    def add_all(self, bulk_credentials, any_cred_eligible=True):
        """
        Add and manage client credentials in bulk.
        """
        for cred in bulk_credentials:
            self.add(cred, any_cred_eligible=any_cred_eligible)

    def remove(self, cred):
        """
        Stop managing the processor associated with the given credentials
        """
        del self._credentials[cred]
        try:
            del self._any_cred_queue[cred]
        except KeyError:
            pass  # Was added with any_cred_eligible=False

    def __getitem__(self, cred):
        """
        Get the processor associated with the given credentials.

        Client code should not store a reference to the processor. You
        should call this method every time to avoid the possibility of
        multiple processors for the same credentials. (The ``maintain``
        method may deallocate some processors if not recently used.)

        :param cred: the client's credentials, or ``ANY_CREDENTIALS`` to
            make promiscuous requests.
        :return: the client's processor or None if a ``ANY_CREDENTIALS``
            request was made with no eligible processors
        """
        if cred is ANY_CREDENTIALS:
            if self._any_cred_queue:
                return self[self._any_cred_queue.pop_and_rotate()]
            else:
                return None
        else:
            managed = self._credentials.get(cred)

            if managed is not None:
                return managed
            else:
                self.add(cred)
                return self._credentials[cred]

    async def maintain(self, cull_interval=FIFTEEN_MINUTES):
        """
        Maintains the request processors, forever.

        Mostly, this means cleaning up request processors, allowing GC.

        :param cull_interval: how frequently, in seconds, maintain
            should run.
        """
        while True:
            await asyncio.sleep(cull_interval)

            try:
                self._maintain()
            except KeyboardInterrupt:
                raise
            except asyncio.CancelledError:
                break
            except Exception as e:
                LOGGER.error("Error %s during maintain()", e)

    def _maintain(self):
        now = time.time()

        # Cleanup all the request processors and identify those which
        # can be removed.
        removal_keys = []
        for k, req_processor in self._credentials.items():
            req_processor.cleanup(now)  # Remove disused rate limits

            if req_processor.is_removable():
                removal_keys.append(k)

        # Remove all those processors not in use or rate limited.
        for k in removal_keys:
            if k not in self._any_cred_queue:
                self.remove(k)


def debug_blocking_request(app_cred, client_cred, twitter_req):
    """
    Fulfill a request using requests (for humans).

    WARNING: This method is for debugging purposes. It does no rate-limit
        enforcement.

    :param app_cred: the AppCredentials
    :param client_cred: the ClientCredentials
    :param twitter_req: The TwitterRequest

    :return: a requests.Response
    """
    # I don't want to require requests as a dependency. And, I really don't
    # want to encourage the use of a blocking function, especially outside
    # the system of rate limit management. But, when debugging, it's terribly
    # useful to have a the following function.
    from requests import request as _request

    headers = generate_req_headers(twitter_req, app_cred, client_cred)

    return _request(twitter_req.method,
                    twitter_req.url,
                    params=twitter_req.params,
                    headers=headers)



import inspect
from importlib import import_module
from brittle_wit_core import TwitterRequest
from brittle_wit import ads_api, rest_api, streaming_api, webhooks_api


# inspect has these methods: XXX
def submodules_of(module):
    parent_name, modules = module.__name__, []

    for name in dir(module):
        if name.startswith('__') and name.endswith('__'):
            continue

        if not inspect.ismodule(getattr(module, name)):
            continue

        if name == 'brittle_wit':
            continue

        path = "{}.{}".format(parent_name, name)

        modules.append(import_module(path))

    return modules


def api_calls_in(module):
    api_calls = []

    for name in dir(module):
        if not name.lower() == name or name.startswith('_'):
            continue
        api_calls.append(getattr(module, name))

    return api_calls


# =============================================================================
# Eventually this needs to be replaced with something more robust. For now,
# it just checks that everything works as in "doesn't err." XXX
# =============================================================================

def test_return_types():
    numbers = list(str(i) for i in range(10))
    for package in [ads_api, rest_api, streaming_api, webhooks_api]:
        for module in submodules_of(package):
            for api_call in api_calls_in(module):
                arg_names = inspect.getargs(api_call.__code__).args
                args = dict(zip(arg_names, numbers[:len(arg_names)]))
                assert isinstance(api_call(**args), TwitterRequest)



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def info_by_stream_id(stream_id):
    """
    Retrieves information about the established stream represented by the
    """
    binding = {'stream_id': stream_id}
    url = 'https://sitestream.twitter.com/2b/site/c/01_225167_334389048B872A533002B34D73F8C29FD09EFC50/info.json'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'streaming:c',
                           'get-c-stream-id-info',
                           binding)


_TwitterRequest.DOC_URLS['https://sitestream.twitter.com/2b/site/c/01_225167_334389048B872A533002B34D73F8C29FD09EFC50/info.json'] = 'https://dev.twitter.com/streaming/reference/get/c/stream_id/info'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################

import brittle_wit.rest_api.search
import brittle_wit.rest_api.collections
import brittle_wit.rest_api.friends
import brittle_wit.rest_api.blocks
import brittle_wit.rest_api.favorites
import brittle_wit.rest_api.application
import brittle_wit.rest_api.geo
import brittle_wit.rest_api.followers
import brittle_wit.rest_api.statuses
import brittle_wit.rest_api.help
import brittle_wit.rest_api.users
import brittle_wit.rest_api.mutes
import brittle_wit.rest_api.saved_searches
import brittle_wit.rest_api.direct_messages
import brittle_wit.rest_api.trends
import brittle_wit.rest_api.lists
import brittle_wit.rest_api.account
import brittle_wit.rest_api.media
import brittle_wit.rest_api.friendships



import pytest

from brittle_wit.app import App
from brittle_wit import AppCredentials, ClientCredentials


@pytest.fixture
def app_cred():
    return AppCredentials('app', 'secret')


@pytest.fixture
def client_cred():
    return ClientCredentials(1, 'user', 'secret')


@pytest.fixture
def app(app_cred):
    app = App.with_defaults(app_cred)
    yield app
    app._session.close()



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def conversion_attribution(app_id, conversion_time, conversion_type,
                           hashed_device_id, os_type, *, click_window=_ELIDE,
                           extra_device_ids=_ELIDE,
                           non_twitter_engagement_time=_ELIDE,
                           non_twitter_engagement_type=_ELIDE,
                           view_through_window=_ELIDE):
    """
    Query Twitter to check on conversion attribution without writing a
    conversion event. Response will indicate Twitter attribution. Relates to

    :param app_id: The unique identifier with the corresponding app store.
        Type: int, string Example: 333903271, com.vine.android

    :param conversion_time: The time of the conversion event in an ISO-8601
        timestamp format, with milliseconds appended. Type: string Example:
        2014-05-22T02:38:28.103Z

    :param conversion_type: The type of conversion event. Type: enum Possible
        values: ACHIEVEMENT_UNLOCKED, ADDED_PAYMENT_INFO, ADD_TO_CART,
        ADD_TO_WISHLIST, CHECKOUT_INITIATED, CONTENT_VIEW, INSTALL, INVITE,
        LEVEL_ACHIEVED, LOGIN, PURCHASE, RATED, RESERVATION, RE_ENGAGE, SEARCH,
        SHARE, SIGN_UP, SPENT_CREDITS, TUTORIAL_COMPLETE, UPDATE

    :param hashed_device_id: The HMAC_SHA-256 hashed IDFA or AdID. Type: string
        Example: ABCD1234XYZ

    :param os_type: The OS type for the app. Type: enum Possible values: IOS,
        ANDROID

    :param click_window: The click window for this event in days. Note: Only
        the possible values listed below are accepted. Type: int Default: 14
        Possible values: 1, 7, 14, 30

    :param extra_device_ids: A SHA256 of the SHA1 of the device ID passed in
        hashed_device_id, plus any additional hashed device IDs. Type: string
        Example: ABCD1234XYZ, DCBA4321XYZ

    :param non_twitter_engagement_time: The time of the last non-twitter
        engagement prior to the conversion. Type: string Example:
        2014-05-22T02:38:28.103Z

    :param non_twitter_engagement_type: The type of non-twitter engagement
        prior to the conversion event. Type: enum Possible values: CLICK, VIEW

    :param view_through_window: The view through window for this event in days.
        Note: Only the possible values listed below are accepted. Type: int
        Default: 1 Possible values: 0, 1, 7, 14, 30
    """
    binding = {'app_id': app_id, 'conversion_time': conversion_time,
               'conversion_type': conversion_type, 'hashed_device_id':
               hashed_device_id, 'os_type': os_type, 'click_window':
               click_window, 'extra_device_ids': extra_device_ids,
               'non_twitter_engagement_time': non_twitter_engagement_time,
               'non_twitter_engagement_type': non_twitter_engagement_type,
               'view_through_window': view_through_window}
    url = 'https://ads-api.twitter.com/2/conversion_attribution'
    return _TwitterRequest('GET',
                           url,
                           'ads:conversion_attribution',
                           'get-conversion-attribution',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/conversion_attribution'] = 'https://dev.twitter.com/ads/reference/get/conversion_attribution'



###############################################################################
# THIS FILE WAS AUTO-GENERATED!
###############################################################################


from brittle_wit_core import TwitterRequest as _TwitterRequest
from brittle_wit_core import ELIDE as _ELIDE


def accounts_account_id(account_id, audience_type, *, audience_values=_ELIDE,
                        interaction_type=_ELIDE):
    """
    Retrieve insights for specific account.

    :param account_id: The identifier for the leveraged account. Appears within
        the resource’s path and is generally a required parameter for all
        Advertiser API requests excluding GET accounts. The specified account
        must be associated with the authenticating user. Type: string Example:
        18ce54d4x5t

    :param audience_type: The audience or set of people you’d like more
        insights on. See Ads Enumerations for more information. Type: enum
        Possible Values: ALL_ON_TWITTER, APP, CAMPAIGN, ORGANIC

    :param audience_values: The identifier (ID) of the available audience set
        requested. Availability depends on the audience_type. Required for
        ORGANIC. Type: string Example: 702

    :param interaction_type: The type of interaction captured for the available
        audience set requested. on the audience_type. Required for ORGANIC.
        Type: enum Possible Values: CONVERSION, ENGAGEMENT, IMPRESSION
    """
    binding = {'account_id': account_id, 'audience_type': audience_type,
               'audience_values': audience_values, 'interaction_type':
               interaction_type}
    url = 'https://ads-api.twitter.com/2/insights/accounts/:account_id'
    return _TwitterRequest('GET',
                           url,
                           'ads:insights',
                           'get-insights-accounts-account-id',
                           binding)


def accounts_available_audiences_by_account_id(account_id):
    """
    Retrieve available audiences for this account.

    :param account_id: The identifier for the leveraged account. Appears within
        the resource’s path and is generally a required parameter for all
        Advertiser API requests excluding GET accounts. The specified account
        must be associated with the authenticating user. Type: string Example:
        18ce54d4x5t
    """
    binding = {'account_id': account_id}
    url = 'https://ads-api.twitter.com/2/insights/accounts/{account_id}/available_audiences'
    url = url.format(**binding)
    return _TwitterRequest('GET',
                           url,
                           'ads:insights',
                           'get-insights-accounts-account-id-available-audiences',
                           binding)


def keywords_search(end_time, granularity, start_time, *, keywords=_ELIDE,
                    location=_ELIDE, negative_keywords=_ELIDE):
    """
    Given a group of keywords, get the associated tweet volume for up to 7
    days.

    :param end_time: Scopes the retrieved data to data collected in the window
        of time between start_time and end_time. If not set, this defaults to
        the time of the request. Expressed in ISO 8601. Type: string Example:
        2015-07-06T07:59:00Z

    :param granularity: Specifies the granularity of the data returned for the
        time range denoted by start_time and end_time. For instance, when set
        to HOUR, you will be presented with a datapoint for each hour between
        start_time and end_time. request. Type: enum Possible values: DAY, HOUR

    :param start_time: Scopes the retrieved data to data collected in the
        window of time between start_time and end_time. Expressed in ISO 8601.
        Type: string Example: 2015-07-06T07:59:00Z

    :param keywords: A comma-separated string of keywords to narrow search by.
        All keywords are OR’ed with one another. A maximum of 10 keywords and
        negative_keywords (combined) may be used. Type: string Example: good
        happy shiny

    :param location: A targeting value you would get from the GET
        targeting_criteria/_locations endpoint to narrow results in terms of
        where the user of the account is located. Note that at present only
        country level locations are supported. Type: string Example:
        0ce8b9a7b2742f7e

    :param negative_keywords: A comma-separated string of keywords to exclude
        (black list). All negative keywords are OR’ed with one another. A
        maximum of 10 keywords and negative_keywords (combined) may be used.
        Type: string Example: horrible bad
    """
    binding = {'end_time': end_time, 'granularity': granularity, 'start_time':
               start_time, 'keywords': keywords, 'location': location,
               'negative_keywords': negative_keywords}
    url = 'https://ads-api.twitter.com/2/insights/keywords/search'
    return _TwitterRequest('GET',
                           url,
                           'ads:insights',
                           'get-insights-keywords-search',
                           binding)


_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/insights/accounts/:account_id'] = 'https://dev.twitter.com/ads/reference/get/insights/accounts/account_id'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/insights/accounts/{account_id}/available_audiences'] = 'https://dev.twitter.com/ads/reference/get/insights/accounts/account_id/available_audiences'
_TwitterRequest.DOC_URLS['https://ads-api.twitter.com/2/insights/keywords/search'] = 'https://dev.twitter.com/ads/reference/get/insights/keywords/search'



#!/usr/bin/env python

import json
import os
from collections import defaultdict
from vaquero import Vaquero, ModulePipeline
from vaquero.util import jsonlines_reader


SELF_DIR = os.path.dirname(os.path.realpath(__file__))
RAW_DIR = os.path.join(SELF_DIR, "data", "raw")
CLEAN_DIR = os.path.join(SELF_DIR, "data", "clean")
INPUT_PATH = os.path.join(RAW_DIR, "raw_html.jsonl")
OUTPUT_PATH = os.path.join(CLEAN_DIR, "api.json")

IGNORE_PAGES = {'https://dev.twitter.com/ads/reference/get',
                'https://dev.twitter.com/ads/reference/post',
                'https://dev.twitter.com/ads/reference/put',
                'https://dev.twitter.com/ads/reference/del'}

if __name__ == '__main__':
    os.makedirs(CLEAN_DIR, exist_ok=True)
    definitions = defaultdict(list)

    vaq = Vaquero(max_failures=0)
    import api_pipeline
    pipeline = ModulePipeline(api_pipeline)

    # These reset so you can do `%run -i extract_definitions.py` in a notebook
    vaq.reset()
    pipeline.reload()

    for src in jsonlines_reader(INPUT_PATH):
        dst = {}
        if src['reference_url'] in IGNORE_PAGES:
            continue
        print("Parsing:", src['reference_url'])
        with vaq:
            pipeline(src, dst)
            definitions[dst['group']].append(dst)

    print(vaq.stats())
    if not os.environ.get('DEBUG'):
        with open(OUTPUT_PATH, "w") as fp:
            json.dump(definitions, fp, indent=4)



import pytest
from brittle_wit.helpers import CircularQueue


@pytest.fixture
def q():
    return CircularQueue()


def test_contains(q):
    assert 'a' not in q
    q.add('a')
    assert 'a' in q


def test_bool(q):
    assert not q
    q.add('a')
    assert q


def test_add_raises_exception_on_already_contained(q):
    q.add('a')
    with pytest.raises(ValueError) as e:
        q.add('a')
    assert "already in queue" in str(e)


def test_del(q):
    q.add('a')
    assert 'a' in q
    del q['a']
    assert 'a' not in q


def test_del_raises_key_error(q):
    q.add('a')
    assert 'a' in q
    del q['a']
    with pytest.raises(KeyError):
        del q['a']


def test_pop_and_rotate(q):
    q.add('a')
    q.add('b')
    items = [q.pop_and_rotate() for _ in range(5)]
    assert items == list('babab')



import pytest
import asyncio
import time
from brittle_wit_core import (AppCredentials, ClientCredentials,
                              TwitterRequest, TwitterResponse, TwitterError)
from brittle_wit.executors import ClientRequestProcessor
from brittle_wit.rate_limit import RateLimit
from tests.helpers import MockHTTPReq, MockHTTPResp


app_cred = AppCredentials("key", "secret")
client_cred = ClientCredentials(2, "token-bob", "****")


@pytest.fixture
def processor():
    return ClientRequestProcessor(client_cred)


def test_properties(processor):
    assert processor.user_id == 2


def test_is_removable_after_init(processor):
    assert processor.is_removable()


def test_is_removable_when_processing(processor, mocker):
    mgr = mocker.MagicMock()
    mgr.all_chains_empty = lambda: False
    processor._chain_manager = mgr

    # It's waiting for a ticket to make a request.
    assert not processor.is_removable()


def test_is_removable_when_rate_limited(processor):
    processor._rate_limits['svc'] = RateLimit.from_initial_limit(0)
    assert not processor.is_removable()


def test_is_not_processing_on_init(processor):
    assert processor.is_not_processing()


def test_is_processing_when_waiting_for_crit_region(processor, mocker):
    mgr = mocker.MagicMock()
    mgr.all_chains_empty = lambda: False
    processor._chain_manager = mgr

    # It's waiting for a ticket to make a request.
    assert not processor.is_not_processing()


def test_is_immediately_available_for_when_no_line_and_no_limit(processor):
    assert processor.is_immediately_available_for('svc')


def test_is_immediately_available_for_when_cur_on_line(processor, mocker):
    mgr = mocker.MagicMock()
    mgr.is_chain_empty = lambda _: False
    processor._chain_manager = mgr

    assert not processor.is_immediately_available_for('svc')


def test_is_immediately_available_for_when_no_line_and_exhausted(processor):
    processor._rate_limits['svc'] = RateLimit.from_initial_limit(0)
    assert not processor.is_immediately_available_for('svc')


def test_is_immediately_available_for_when_no_line_and_not_exhaust(processor):
    processor._rate_limits['svc'] = RateLimit.from_ignorance()
    assert processor.is_immediately_available_for('svc')


def test_is_immediately_available_for_when_line_not_empty(processor, mocker):
    svc = "direct_message"
    assert processor.is_immediately_available_for(svc)

    limit = RateLimit.from_ignorance()
    limit._remaining = 0
    processor._rate_limits[svc] = limit
    assert not processor.is_immediately_available_for(svc)

    tm = mocker.MagicMock()
    tm.is_line_empty = lambda s: False
    processor._ticket_master = tm
    assert not processor.is_immediately_available_for(svc)


def test_rate_limit_for_doesnt_duplicate_on_same_service(processor):
    limit = processor._rate_limit_for("direct_messages")
    assert isinstance(limit, RateLimit)  # Sane

    assert limit is processor._rate_limit_for("direct_messages")


def test_rate_limit_for_unique_for_diff_service(processor):
    dm_limit = processor._rate_limit_for("direct_messages")
    settings_limit = processor._rate_limit_for("account/settings")

    assert isinstance(dm_limit, RateLimit)  # Sane
    assert isinstance(settings_limit, RateLimit)  # Sane.
    assert dm_limit is not settings_limit


def test_rate_limit_for_is_initially_prestine(processor):
    limit = processor._rate_limit_for("direct_messages")
    assert limit.is_pristine


def test_cleanup_with_no_rate_limits(processor):
    assert not processor._rate_limits
    processor.cleanup()
    assert not processor._rate_limits


def test_cleanup_with_stale_rate_limit(processor):
    now = time.time()
    processor._rate_limits['past'] = RateLimit(1, 1, now - 100)
    processor.cleanup(now)
    assert processor.is_removable()


def test_cleanup_with_mixed_rate_limits(processor):
    now = time.time()
    processor._rate_limits['future'] = RateLimit(1, 1, now + 100)
    processor._rate_limits['past'] = RateLimit(1, 1, now - 100)
    assert set(processor._rate_limits), {'future', 'past'}

    processor.cleanup(now)

    assert not processor.is_removable()
    assert set(processor._rate_limits), {'future'}


class RateLimitKludge:
    # XXX: TODO: REFACTOR
    # This is a kludge that lets me test the execute.
    # It's extremely implementation dependent, and generally a shit idea.

    def __init__(self, twitter_req, client_cred=None, fail_queue=None, sleep_time=None):
        self.twitter_req = twitter_req
        self.client_cred = client_cred
        self.fail_queue = fail_queue
        self.sleep_time = sleep_time

    async def comply(self):
        await asyncio.sleep(10)

    def update(self, _):
        pass

    async def comply(self):
        if self.fail_queue is not None and self.fail_queue.pop(0):
            resp = MockHTTPResp(500, {}, "")
            raise TwitterError(self.client_cred, self.twitter_req, resp, "")

        if self.sleep_time is not None:
            await asyncio.sleep(self.sleep_time)


@pytest.mark.asyncio
async def test__execute_prestine(processor, mocker):
    rate_limit = RateLimit.from_ignorance()
    processor = ClientRequestProcessor(client_cred)
    http_req = MockHTTPReq(MockHTTPResp(200, {}, {'count': 100}))

    twitter_req = TwitterRequest("method", "url", "family", "service")
    coro = processor._execute('session', app_cred, twitter_req)
    func_path = 'brittle_wit.executors.twitter_req_to_http_req'
    req_to_http_req = mocker.patch(func_path, return_value=http_req)

    resp = await coro

    req_to_http_req.assert_called_with('session',
                                       app_cred,
                                       client_cred,
                                       twitter_req)

    assert type(resp) is TwitterResponse
    assert resp.body['count'] == 100


@pytest.mark.asyncio
async def test__execute_timeout_thrown(processor):
    twitter_req = TwitterRequest("method", "url", "family", "service")
    coro = processor._execute('session', app_cred, twitter_req, 0.01)
    processor._rate_limits['service'] = RateLimitKludge(twitter_req, client_cred, sleep_time=1)

    with pytest.raises(asyncio.TimeoutError):
        await coro


@pytest.mark.asyncio
async def test_execute_retrying(processor, mocker):
    twitter_req = TwitterRequest("method", "url", "family", "service")
    http_req = MockHTTPReq(MockHTTPResp(200, {}, {'count': 100}))
    fail_queue = [True, False]

    processor._rate_limits['service'] = RateLimitKludge(twitter_req, client_cred, fail_queue)
    func_path = 'brittle_wit.executors.twitter_req_to_http_req'
    coro = processor.execute('session', app_cred, twitter_req, sleep_time=0.01)
    req_to_http_req = mocker.patch(func_path, return_value=http_req)
    resp = await coro
    assert type(resp) is TwitterResponse
    assert resp.body['count'] == 100
    assert fail_queue == []


@pytest.mark.asyncio
async def test_execute_retrying_too_many_failures(processor, mocker):
    twitter_req = TwitterRequest("method", "url", "family", "service")
    http_req = MockHTTPReq(MockHTTPResp(200, {}, {'count': 100}))
    fail_queue = [True] * 6

    processor._rate_limits['service'] = RateLimitKludge(twitter_req, client_cred, fail_queue)
    func_path = 'brittle_wit.executors.twitter_req_to_http_req'
    coro = processor.execute('session', app_cred, twitter_req, sleep_time=0.01)
    req_to_http_req = mocker.patch(func_path, return_value=http_req)
    with pytest.raises(TwitterError):
        resp = await coro
